import tensorflow as tf

from neupy.utils import asfloat
from neupy.layers.utils import iter_parameters
from neupy.core.properties import NumberProperty
from .base import WeightUpdateConfigurable


__all__ = ('MaxNormRegularization',)


def max_norm_clip(array, max_norm):
    """
    Clips norm of the array in case if it will exceed
    specified limit.

    Parameters
    ----------
    array : array-like
        Any array-like object.

    max_norm : float
        Maximum possible norm value for the input array. Array will
        be clipped in case if its norm will be greater than
        specified limit.
    """
    with tf.name_scope('max-norm-clip'):
        array_norm = tf.norm(array)
        max_norm = asfloat(max_norm)

        return tf.where(
            tf.greater_equal(array_norm, max_norm),
            tf.multiply(max_norm, array) / array_norm,
            array,
        )


class MaxNormRegularization(WeightUpdateConfigurable):
    """
    Max-norm regularization algorithm will clip norm of the
    parameter in case if it will exceed maximum limit.

    .. code-block:: python

        if norm(weight) > max_norm:
            weight = max_norm * weight / norm(weight)

    .. raw:: html

        <br>

    Warns
    -----
    {WeightUpdateConfigurable.Warns}

    Parameters
    ----------
    max_norm : int, float
        Any parameter that has norm greater than this value
        will be clipped. Defaults to ``10``.

    Examples
    --------
    >>> from neupy import algorithms
    >>> bpnet = algorithms.GradientDescent(
    ...     (2, 4, 1),
    ...     step=0.1,
    ...     max_norm=4,
    ...     addons=[algorithms.MaxNormRegularization]
    ... )

    References
    ----------
    [1] N. Srivastava, G. Hinton, A. Krizhevsky, I. Sutskever,
      R. Salakhutdinov. Dropout: A Simple Way to Prevent
      Neural Networks from Overfitting.
      http://jmlr.org/papers/volume15/srivastava14a/srivastava14a.pdf
    """
    max_norm = NumberProperty(default=10, minval=0)

    def init_train_updates(self):
        parameters = [param for _, _, param in iter_parameters(self.layers)]
        original_updates = super(
            MaxNormRegularization, self).init_train_updates()

        modified_updates = []

        for parameter, updated in original_updates:
            if parameter in parameters:
                updated = max_norm_clip(updated, self.max_norm)
            modified_updates.append((parameter, updated))

        return modified_updates



import tensorflow as tf

from neupy import init
from neupy.utils import as_tuple
from neupy.core.properties import IntProperty, ParameterProperty
from .base import BaseLayer


__all__ = ('Embedding',)


class Embedding(BaseLayer):
    """
    Embedding layer accepts indeces as an input and returns
    rows from the weight matrix associated with these indeces.
    Useful in case of categorical features or for the word
    embedding tasks.

    Parameters
    ----------
    input_size : int
        Layer's input vector dimension. Usualy associated with number
        of categories or number of unique words that input vector has.

    output_size : int
        Layer's output vector dimension.

    weight : array-like, Tensorfow variable, scalar or Initializer
        Defines layer's weights. Default initialization methods
        you can find :ref:`here <init-methods>`.
        Defaults to :class:`XavierNormal() <neupy.init.XavierNormal>`.

    {BaseLayer.Parameters}

    Methods
    -------
    {BaseLayer.Methods}

    Attributes
    ----------
    {BaseLayer.Attributes}

    Examples
    --------

    This example converts dataset that has only categorical
    variables into format that suitable for Embedding layer.

    >>> import numpy as np
    >>> from neupy import layers
    >>>
    >>> dataset = np.array([
    ...     ['cold', 'high'],
    ...     ['hot',  'low'],
    ...     ['cold', 'low'],
    ...     ['hot',  'low'],
    ... ])
    >>>
    >>> unique_value, dataset_indeces = np.unique(
    ...     dataset, return_inverse=True
    ... )
    >>> dataset_indeces = dataset_indeces.reshape((4, 2))
    >>> dataset_indeces
    array([[0, 1],
           [2, 3],
           [0, 3],
           [2, 3]])
    >>>
    >>> n_features = dataset.shape[1]
    >>> n_unique_categories = len(unique_value)
    >>> embedded_size = 1
    >>>
    >>> connection = layers.join(
    ...     layers.Input(n_features),
    ...     layers.Embedding(n_unique_categories, embedded_size),
    ...     # Output from the embedding layer is 3D
    ...     # To make output 2D we need to reshape dimensions
    ...     layers.Reshape(),
    ... )
    """
    input_size = IntProperty(minval=1)
    output_size = IntProperty(minval=1)
    weight = ParameterProperty(default=init.XavierNormal())

    def __init__(self, input_size, output_size, **options):
        super(Embedding, self).__init__(
            input_size=input_size, output_size=output_size, **options)

    @property
    def output_shape(self):
        if self.input_shape is not None:
            return as_tuple(self.input_shape, self.output_size)

    def initialize(self):
        super(Embedding, self).initialize()
        self.add_parameter(
            value=self.weight, name='weight',
            shape=as_tuple(self.input_size, self.output_size),
            trainable=True)

    def output(self, input_value):
        input_value = tf.cast(input_value, tf.int32)
        return tf.gather(self.weight, input_value)

    def __repr__(self):
        classname = self.__class__.__name__
        return '{name}({input_size}, {output_size})'.format(
            name=classname,
            input_size=self.input_size,
            output_size=self.output_size,
        )



from random import randint

import numpy as np

from neupy.utils import format_data
from neupy.exceptions import NotTrained
from .utils import bin2sign, hopfield_energy, step_function
from .base import DiscreteMemory


__all__ = ('DiscreteBAM',)


class DiscreteBAM(DiscreteMemory):
    """
    Discrete BAM Network with associations.
    Network associate every input with some target value.

    Parameters
    ----------
    {DiscreteMemory.mode}

    {DiscreteMemory.n_times}

    Methods
    -------
    train(input_data, output_data)
        Train network and update network weights.

    predict_output(input_data, n_times=None)
        Using input data recover output data. Returns two arguments.
        First is an input data, second is an output data.

    predict(input_data, n_times=None)
        Alias to the ``predict_output`` method.

    predict_input(output_data, n_times=None)
        Using output data recover input data. Returns two arguments.
        First is input data, second is output data.

    energy(input_data, output_data)
        Calculate Hopfield Energy for the input and output data.

    Notes
    -----
    - Input and output vectors should contain only binary values.

    Examples
    --------
    >>> import numpy as np
    >>> from neupy import algorithms
    >>>
    >>> zero = np.matrix([
    ...     0, 1, 1, 1, 0,
    ...     1, 0, 0, 0, 1,
    ...     1, 0, 0, 0, 1,
    ...     1, 0, 0, 0, 1,
    ...     1, 0, 0, 0, 1,
    ...     0, 1, 1, 1, 0
    ... ])
    >>>
    >>> one = np.matrix([
    ...     0, 1, 1, 0, 0,
    ...     0, 0, 1, 0, 0,
    ...     0, 0, 1, 0, 0,
    ...     0, 0, 1, 0, 0,
    ...     0, 0, 1, 0, 0,
    ...     0, 0, 1, 0, 0
    ... ])
    >>> zero.reshape((5, 6))
    >>>
    >>> half_zero = np.matrix([
    ...     0, 1, 1, 1, 0,
    ...     1, 0, 0, 0, 1,
    ...     1, 0, 0, 0, 1,
    ...     0, 0, 0, 0, 0,
    ...     0, 0, 0, 0, 0,
    ...     0, 0, 0, 0, 0,
    ... ])
    >>>
    >>> zero_hint = np.matrix([[0, 1, 0, 0]])
    >>> one_hint = np.matrix([[1, 0, 0, 0]])
    >>>
    >>> data = np.concatenate([zero, one], axis=0)
    >>> hints = np.concatenate([zero_hint, one_hint], axis=0)
    >>>
    >>> bamnet = algorithms.DiscreteBAM(mode='sync')
    >>> bamnet.train(data, hints)
    >>>
    >>> recovered_zero, recovered_hint = bamnet.predict(half_zero)
    >>> recovered_hint
    matrix([[0, 1, 0, 0]])
    >>>
    >>> zero_hint
    matrix([[0, 1, 0, 0]])
    """

    def format_predict(self, predicted_result):
        return step_function(predicted_result).astype(int)

    def predict_input(self, output_data, n_times=None):
        output_data = format_data(output_data, is_feature1d=False)
        return self.prediction(input_data=None, output_data=output_data,
                               n_times=n_times)

    def predict_output(self, input_data, n_times=None):
        input_data = format_data(input_data, is_feature1d=False)
        return self.prediction(input_data=input_data, output_data=None,
                               n_times=n_times)

    def prediction(self, input_data=None, output_data=None, n_times=None):
        if self.weight is None:
            raise NotTrained("Network hasn't been trained yet")

        if input_data is None and output_data is not None:
            self.discrete_validation(output_data)
            output_data = bin2sign(output_data)
            input_data = np.sign(output_data.dot(self.weight.T))

        elif input_data is not None and output_data is None:
            self.discrete_validation(input_data)
            input_data = bin2sign(input_data)
            output_data = np.sign(input_data.dot(self.weight))

        else:
            raise ValueError("Prediction is possible only for input or output")

        n_output_features = output_data.shape[-1]
        n_input_features = input_data.shape[-1]

        if self.mode == 'async':
            if n_times is None:
                n_times = self.n_times

            for _ in range(n_times):
                input_position = randint(0, n_input_features - 1)
                output_position = randint(0, n_output_features - 1)

                input_data[:, input_position] = np.sign(
                    output_data.dot(self.weight[input_position, :])
                )
                output_data[:, output_position] = np.sign(
                    input_data.dot(self.weight[:, output_position])
                )

        return (
            self.format_predict(input_data),
            self.format_predict(output_data),
        )

    def train(self, input_data, output_data):
        self.discrete_validation(input_data)
        self.discrete_validation(output_data)

        output_data = bin2sign(format_data(output_data, is_feature1d=False))
        input_data = bin2sign(format_data(input_data, is_feature1d=False))

        _, wight_nrows = input_data.shape
        _, wight_ncols = output_data.shape
        weight_shape = (wight_nrows, wight_ncols)

        if self.weight is None:
            self.weight = np.zeros(weight_shape)

        if self.weight.shape != weight_shape:
            raise ValueError("Invalid input shapes. Number of input "
                             "features must be equal to {} and {} output "
                             "features".format(wight_nrows, wight_ncols))

        self.weight += input_data.T.dot(output_data)

    def energy(self, input_data, output_data):
        self.discrete_validation(input_data)
        self.discrete_validation(output_data)

        input_data, output_data = bin2sign(input_data), bin2sign(output_data)
        input_data = format_data(input_data, is_feature1d=False)
        output_data = format_data(output_data, is_feature1d=False)
        nrows, n_features = input_data.shape

        if nrows == 1:
            return hopfield_energy(self.weight, input_data, output_data)

        output = np.zeros(nrows)
        for i, rows in enumerate(zip(input_data, output_data)):
            output[i] = hopfield_energy(self.weight, *rows)

        return output

    predict = predict_output



import numpy as np
from numpy.core.umath_tests import inner1d


__all__ = ('bin2sign', 'hopfield_energy', 'step_function')


def bin2sign(matrix):
    """
    Convert zeros in banary matrix to -1 values.

    Parameters
    ----------
    matrix : array-like

    Returns
    -------
    array-like
        Return they same matrix with modified values.
    """
    return np.where(matrix == 0, -1, 1)


def hopfield_energy(weight, input_data, output_data):
    """
    Compute Hopfield energy between input data, output data and
    neural network weights.

    Parameters
    ----------
    input_data : vector
        Input dataset

    output_data : vector
        Output dataset

    weight : 2D array
        Network's weights.

    Returns
    -------
    float
        Hopfield energy for specific data and weights.
    """
    return -0.5 * inner1d(input_data.dot(weight), output_data)


def step_function(input_value):
    """
    Step function.

    Parameters
    ----------
    input_value : array-like

    Returns
    -------
    array-like
    """
    return np.where(input_value > 0, 1, 0)



import numpy as np
import tensorflow as tf

from neupy.utils import as_tuple
from neupy.exceptions import LayerConnectionError
from neupy.core.properties import TypedListProperty
from .base import BaseLayer


__all__ = ('Reshape', 'Transpose')


class NewShapeProperty(TypedListProperty):
    def validate(self, value):
        super(NewShapeProperty, self).validate(value)

        if value.count(-1) >= 2:
            raise ValueError("Only single -1 value can be specified")


class Reshape(BaseLayer):
    """
    Reshapes input tensor.

    Parameters
    ----------
    shape : tuple or list
        New feature shape. The ``-1`` value means that this value
        will be computed from the total size that remains. If you need
        to get the output feature with more that 2 dimensions then you can
        set up new feature shape using tuples or list. Defaults to ``[-1]``.

    {BaseLayer.Parameters}

    Methods
    -------
    {BaseLayer.Methods}

    Attributes
    ----------
    {BaseLayer.Attributes}

    Examples
    --------

    Covert 4D input to 2D

    >>> from neupy.layers import *
    >>> conn = Input((2, 5, 5)) > Reshape()
    >>> conn.input_shape
    (2, 5, 5)
    >>> conn.output_shape
    (50,)

    Convert 3D to 4D

    >>> from neupy.layers import *
    >>> conn = Input((5, 4)) > Reshape((5, 2, 2))
    >>> conn.input_shape
    (5, 4)
    >>> conn.output_shape
    (5, 2, 2)
    """
    shape = NewShapeProperty(default=(-1,))

    def __init__(self, shape=(-1,), **options):
        super(Reshape, self).__init__(shape=shape, **options)

    @property
    def output_shape(self):
        if -1 not in self.shape:
            return as_tuple(self.shape)

        if None not in self.input_shape:
            known_shape_values = [val for val in self.shape if val != -1]

            flatten_shape = np.prod(self.input_shape)
            expected_shape_parts = np.prod(known_shape_values)

            if flatten_shape % expected_shape_parts != 0:
                raise ValueError(
                    "Cannot derive values for shape {} from the input "
                    "shape {}".format(self.shape, self.input_shape))

            missing_value = int(flatten_shape // expected_shape_parts)
        else:
            missing_value = None

        return as_tuple([
            missing_value if val == -1 else val for val in self.shape])

    def output(self, input_value):
        """
        Reshape the feature space for the input value.

        Parameters
        ----------
        input_value : array-like or Tensorfow variable
        """
        input_shape = tf.shape(input_value)
        n_samples = input_shape[0]
        output_shape = as_tuple(n_samples, self.shape)
        return tf.reshape(input_value, output_shape)


class Transpose(BaseLayer):
    """
    Transposes input. Permutes the dimensions according to ``perm``.

    Parameters
    ----------
    perm : tuple or list
        A permutation of the dimensions of the input tensor. Layer cannot
        transpose batch dimension and using ``0`` in the list of
        permuted dimensions is not allowed.

    {BaseLayer.Parameters}

    Methods
    -------
    {BaseLayer.Methods}

    Attributes
    ----------
    {BaseLayer.Attributes}

    Examples
    --------
    >>> from neupy.layers import *
    >>> conn = Input((7, 11)) > Transpose((2, 1))
    >>> conn.input_shape
    (7, 11)
    >>> conn.output_shape
    (11, 7)
    """
    perm = TypedListProperty()

    def __init__(self, perm, **options):
        if 0 in perm:
            raise ValueError(
                "Batch dimension has fixed position and 0 "
                "index cannot be used.")

        super(Transpose, self).__init__(perm=perm, **options)

    def validate(self, input_shape):
        if len(input_shape) < 2:
            raise LayerConnectionError(
                "Transpose expects input with at least 3 dimensions.")

    @property
    def output_shape(self):
        # Input shape doesn't have information about the batch size and perm
        # indeces require to have this dimension on zero's position.
        input_shape = np.array(as_tuple(None, self.input_shape))
        return as_tuple(input_shape[self.perm].tolist())

    def output(self, input_value):
        # Input value has batch dimension, but perm will never have it
        # specified as (zero index), so we need to add it in order to
        # fix batch dimesnion in place.
        return tf.transpose(input_value, [0] + list(self.perm))



from __future__ import division

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from sklearn import datasets
from neupy import algorithms, environment


environment.reproducible()

GRID_HEIGHT = 40
GRID_WIDTH = 40

digits = datasets.load_digits()
data = digits.data

sofm = algorithms.SOFM(
    n_inputs=64,
    features_grid=(GRID_HEIGHT, GRID_WIDTH),

    # Learning radius defines area within which we find
    # winning neuron neighbours. The higher the value
    # the more values we will be updated after each iteration.
    learning_radius=5,
    # Every 20 epochs learning radius will be reduced by 1.
    reduce_radius_after=20,

    step=0.5,
    std=1,

    shuffle_data=True,
    verbose=True,
)

sofm.train(data, epochs=100)
clusters = sofm.predict(data).argmax(axis=1)

print("Building visualization...")
plt.figure(figsize=(12, 11))

grid = gridspec.GridSpec(GRID_HEIGHT, GRID_WIDTH)
grid.update(wspace=0, hspace=0)

for row_id in range(GRID_HEIGHT):
    print("Progress: {:.2%}".format(row_id / GRID_HEIGHT))

    for col_id in range(GRID_WIDTH):
        index = row_id * GRID_HEIGHT + col_id
        clustered_samples = data[clusters == index]

        if len(clustered_samples) > 0:
            # We take the first sample, but it can be any
            # sample from this cluster
            sample = clustered_samples[0]

        else:
            # If we don't have samples in cluster then
            # it means that there is a gap in space
            sample = np.zeros(64)

        plt.subplot(grid[index])
        plt.imshow(sample.reshape((8, 8)), cmap='Greys')
        plt.axis('off')

print("Visualization has been built succesfully")
plt.show()



from __future__ import division

import numpy as np
from numpy.linalg import norm

from neupy.utils import format_data
from neupy.core.properties import IntProperty, WithdrawProperty
from neupy.algorithms.gd import StepSelectionBuiltIn
from neupy.algorithms.base import BaseNetwork


__all__ = ('RBFKMeans',)


class RBFKMeans(StepSelectionBuiltIn, BaseNetwork):
    """
    Radial basis function K-means for clustering.

    Parameters
    ----------
    n_clusters : int
        Number of clusters.

    {BaseNetwork.show_epoch}

    {BaseNetwork.shuffle_data}

    {BaseNetwork.epoch_end_signal}

    {BaseNetwork.train_end_signal}

    {Verbose.verbose}

    Attributes
    ----------
    centers : array-like with shape (n_clusters, n_futures)
        Cluster centers.

    Methods
    -------
    train(input_train, epsilon=1e-5, epochs=100)
        Trains network.

    {BaseSkeleton.predict}

    {BaseSkeleton.fit}

    Examples
    --------
    >>> import numpy as np
    >>> from neupy.algorithms import RBFKMeans
    >>>
    >>> data = np.array([
    ...     [0.11, 0.20],
    ...     [0.25, 0.32],
    ...     [0.64, 0.60],
    ...     [0.12, 0.42],
    ...     [0.70, 0.73],
    ...     [0.30, 0.27],
    ...     [0.43, 0.81],
    ...     [0.44, 0.87],
    ...     [0.12, 0.92],
    ...     [0.56, 0.67],
    ...     [0.36, 0.35],
    ... ])
    >>> rbfk_net = RBFKMeans(n_clusters=2, verbose=False)
    >>> rbfk_net.train(data, epsilon=1e-5)
    >>> rbfk_net.centers
    array([[ 0.228     ,  0.312     ],
           [ 0.48166667,  0.76666667]])
    >>>
    >>> new_data = np.array([[0.1, 0.1], [0.9, 0.9]])
    >>> rbfk_net.predict(new_data)
    array([[ 0.],
           [ 1.]])
    """
    n_clusters = IntProperty(minval=2)
    step = WithdrawProperty()

    def __init__(self, **options):
        self.centers = None
        super(RBFKMeans, self).__init__(**options)

    def predict(self, input_data):
        input_data = format_data(input_data)

        centers = self.centers
        classes = np.zeros((input_data.shape[0], 1))

        for i, value in enumerate(input_data):
            classes[i] = np.argmin(norm(centers - value, axis=1))

        return classes

    def train_epoch(self, input_train, target_train):
        centers = self.centers
        old_centers = centers.copy()
        output_train = self.predict(input_train)

        for i, center in enumerate(centers):
            positions = np.argwhere(output_train[:, 0] == i)

            if not np.any(positions):
                continue

            class_data = np.take(input_train, positions, axis=0)
            centers[i, :] = (1 / len(class_data)) * np.sum(class_data, axis=0)

        return np.abs(old_centers - centers)

    def train(self, input_train, epsilon=1e-5, epochs=100):
        n_clusters = self.n_clusters
        input_train = format_data(input_train)
        n_samples = input_train.shape[0]

        if n_samples <= n_clusters:
            raise ValueError("Number of samples in the dataset is less than "
                             "spcified number of clusters. Got {} samples, "
                             "expected at least {} (for {} clusters)"
                             "".format(n_samples, n_clusters + 1, n_clusters))

        self.centers = input_train[:n_clusters, :].copy()
        super(RBFKMeans, self).train(input_train, epsilon=epsilon,
                                     epochs=epochs)



import tensorflow as tf

from neupy.utils import asfloat
from neupy.core.properties import ProperFractionProperty, NumberProperty
from .base import MinibatchGradientDescent


__all__ = ('Adamax',)


class Adamax(MinibatchGradientDescent):
    """
    AdaMax algorithm.

    Parameters
    ----------
    beta1 : float
        Decay rate. Value need to be between ``0`` and ``1``.
        Defaults to ``0.9``.

    beta2 : float
        Decay rate. Value need to be between ``0`` and ``1``.
        Defaults to ``0.999``.

    epsilon : float
        Value need to be greater than ``0``. Defaults to ``1e-7``.

    step : float
        Learning rate, defaults to ``0.002``.

    {MinibatchGradientDescent.batch_size}

    {GradientDescent.addons}

    {ConstructibleNetwork.connection}

    {ConstructibleNetwork.error}

    {BaseNetwork.show_epoch}

    {BaseNetwork.shuffle_data}

    {BaseNetwork.epoch_end_signal}

    {BaseNetwork.train_end_signal}

    {Verbose.verbose}

    Attributes
    ----------
    {MinibatchGradientDescent.Attributes}

    Methods
    -------
    {MinibatchGradientDescent.Methods}

    Examples
    --------
    >>> import numpy as np
    >>> from neupy import algorithms
    >>>
    >>> x_train = np.array([[1, 2], [3, 4]])
    >>> y_train = np.array([[1], [0]])
    >>>
    >>> mnet = algorithms.Adamax((2, 3, 1))
    >>> mnet.train(x_train, y_train)

    References
    ----------
    [1] Diederik P. Kingma, Jimmy Lei Ba
        Adam: a Method for Stochastic Optimization.
        https://arxiv.org/pdf/1412.6980.pdf
    """
    step = NumberProperty(default=0.002, minval=0)
    beta1 = ProperFractionProperty(default=0.9)
    beta2 = ProperFractionProperty(default=0.999)
    epsilon = NumberProperty(default=1e-7, minval=0)

    def init_variables(self):
        super(Adamax, self).init_variables()

        self.variables.iteration = tf.Variable(
            asfloat(1),
            name='iteration',
            dtype=tf.float32,
        )

    def init_train_updates(self):
        updates = []

        iteration = self.variables.iteration
        step = self.variables.step
        beta1 = self.beta1
        beta2 = self.beta2

        scale = step / (1. - beta1 ** iteration)

        for layer, parameter, gradient in self.iter_params_and_grads():
            prev_first_moment = tf.Variable(
                tf.zeros(parameter.shape),
                name="{}/prev-first-moment".format(parameter.op.name),
                dtype=tf.float32,
            )
            prev_weighted_inf_norm = tf.Variable(
                tf.zeros(parameter.shape),
                name="{}/prev-weighted-inf-norm".format(parameter.op.name),
                dtype=tf.float32,
            )

            first_moment = beta1 * prev_first_moment + (1. - beta1) * gradient
            weighted_inf_norm = tf.maximum(
                beta2 * prev_weighted_inf_norm,
                tf.abs(gradient),
            )

            parameter_delta = (
                scale * (first_moment / (weighted_inf_norm + self.epsilon)))

            updates.extend([
                (prev_first_moment, first_moment),
                (prev_weighted_inf_norm, weighted_inf_norm),
                (parameter, parameter - parameter_delta),
            ])

        updates.append((iteration, iteration + 1))
        return updates



import numpy as np
import tensorflow as tf

from neupy.utils import tensorflow_session, flatten, function_name_scope
from neupy.core.properties import (BoundedProperty, ChoiceProperty,
                                   WithdrawProperty)
from neupy.algorithms import GradientDescent
from neupy.algorithms.gd import StepSelectionBuiltIn, errors
from neupy.algorithms.utils import (parameter_values, setup_parameter_updates,
                                    make_single_vector)


__all__ = ('LevenbergMarquardt',)


@function_name_scope
def compute_jacobian(values, parameters):
    """
    Compute jacobian.

    Parameters
    ----------
    values : Tensorfow variable
        Computed MSE for each sample separetly.

    parameters : list of Tensorfow variable
        Neural network parameters (e.g. weights, biases).

    Returns
    -------
    Tensorfow variable
    """
    values_shape = tf.shape(values)
    n_samples = values_shape[0]

    def compute_gradient_per_value(index, result):
        gradients = tf.gradients(values[index], parameters)
        full_gradient = make_single_vector(gradients)
        return (index + 1, result.write(index, full_gradient))

    _, jacobian = tf.while_loop(
        lambda index, _: index < n_samples,
        compute_gradient_per_value,
        [
            tf.constant(0, tf.int32),
            tf.TensorArray(tf.float32, size=n_samples),
        ]
    )

    return jacobian.stack()


class LevenbergMarquardt(StepSelectionBuiltIn, GradientDescent):
    """
    Levenberg-Marquardt algorithm is a variation of the Newton's method.
    It minimizes MSE error. The algorithm approximates Hessian matrix using
    dot product between two jacobian matrices.

    Notes
    -----
    - Method requires all training data during propagation, which means
      it's not allowed to use mini-batches.

    - Network minimizes only Mean Squared Error (MSE) loss function.

    - Efficient for small training datasets, because it
      computes gradient per each sample separately.

    - Efficient for small-sized networks.

    Parameters
    ----------
    {GradientDescent.connection}

    mu : float
        Control invertion for J.T * J matrix, defaults to ``0.1``.

    mu_update_factor : float
        Factor to decrease the mu if update decrese the error, otherwise
        increse mu by the same factor. Defaults to ``1.2``

    error : {{``mse``}}
        Levenberg-Marquardt works only for quadratic functions.
        Defaults to ``mse``.

    {GradientDescent.show_epoch}

    {GradientDescent.shuffle_data}

    {GradientDescent.epoch_end_signal}

    {GradientDescent.train_end_signal}

    {GradientDescent.verbose}

    {GradientDescent.addons}

    Attributes
    ----------
    {GradientDescent.Attributes}

    Methods
    -------
    {GradientDescent.Methods}

    Examples
    --------
    >>> import numpy as np
    >>> from neupy import algorithms
    >>>
    >>> x_train = np.array([[1, 2], [3, 4]])
    >>> y_train = np.array([[1], [0]])
    >>>
    >>> lmnet = algorithms.LevenbergMarquardt((2, 3, 1))
    >>> lmnet.train(x_train, y_train)

    See Also
    --------
    :network:`GradientDescent` : GradientDescent algorithm.
    """
    mu = BoundedProperty(default=0.01, minval=0)
    mu_update_factor = BoundedProperty(default=1.2, minval=1)
    error = ChoiceProperty(default='mse', choices={'mse': errors.mse})

    step = WithdrawProperty()

    def init_variables(self):
        super(LevenbergMarquardt, self).init_variables()
        self.variables.update(
            mu=tf.Variable(self.mu, name='lev-marq/mu'),
            last_error=tf.Variable(np.nan, name='lev-marq/last-error'),
        )

    def init_train_updates(self):
        network_output = self.variables.network_output
        prediction_func = self.variables.train_prediction_func
        last_error = self.variables.last_error
        error_func = self.variables.error_func
        mu = self.variables.mu

        new_mu = tf.where(
            tf.less(last_error, error_func),
            mu * self.mu_update_factor,
            mu / self.mu_update_factor,
        )

        err_for_each_sample = flatten((network_output - prediction_func) ** 2)

        params = parameter_values(self.connection)
        param_vector = make_single_vector(params)

        J = compute_jacobian(err_for_each_sample, params)
        J_T = tf.transpose(J)
        n_params = J.shape[1]

        parameter_update = tf.matrix_solve(
            tf.matmul(J_T, J) + new_mu * tf.eye(n_params.value),
            tf.matmul(J_T, tf.expand_dims(err_for_each_sample, 1))
        )
        updated_params = param_vector - flatten(parameter_update)

        updates = [(mu, new_mu)]
        parameter_updates = setup_parameter_updates(params, updated_params)
        updates.extend(parameter_updates)

        return updates

    def on_epoch_start_update(self, epoch):
        super(LevenbergMarquardt, self).on_epoch_start_update(epoch)

        last_error = self.errors.last()
        if last_error is not None:
            self.variables.last_error.load(last_error, tensorflow_session())



from numpy import dot

from .base import BaseStepAssociative


__all__ = ('Instar',)


class Instar(BaseStepAssociative):
    """
    Instar is a simple unsupervised Neural Network algortihm
    which detects associations.

    Parameters
    ----------
    {BaseAssociative.n_inputs}

    {BaseAssociative.n_outputs}

    {BaseStepAssociative.n_unconditioned}

    {BaseAssociative.weight}

    {BaseStepAssociative.bias}

    {BaseNetwork.step}

    {BaseNetwork.show_epoch}

    {BaseNetwork.shuffle_data}

    {BaseNetwork.epoch_end_signal}

    {BaseNetwork.train_end_signal}

    {Verbose.verbose}

    Methods
    -------
    {BaseSkeleton.predict}

    {BaseAssociative.train}

    {BaseSkeleton.fit}

    Examples
    --------
    >>> import numpy as np
    >>> from neupy import algorithms
    >>>
    >>> train_data = np.array([
    ...     [0, 1, 0, 0],
    ...     [1, 1, 0, 0],
    ... ])
    >>> test_cases = np.array([
    ...     [0, 1, 0, 0],
    ...     [0, 0, 0, 0],
    ...     [0, 0, 1, 1],
    ... ])
    >>>
    >>> instnet = algorithms.Instar(
    ...     n_inputs=4,
    ...     n_outputs=1,
    ...     n_unconditioned=1,
    ...     step=1,
    ...     verbose=False,
    ... )
    >>>
    >>> instnet.train(train_data, epochs=2)
    >>> instnet.predict(test_cases)
    array([[1],
           [0],
           [0]])
    """
    def weight_delta(self, input_row, layer_output):
        n_unconditioned = self.n_unconditioned
        weight = self.weight[n_unconditioned:, :]
        return self.step * dot(
            (input_row[:, n_unconditioned:].T - weight),
            layer_output.T
        )



from neupy.core.config import Configurable
from neupy.algorithms.gd import WEIGHT_PENALTY


__all__ = ('WeightUpdateConfigurable',)


class WeightUpdateConfigurable(Configurable):
    """
    Configuration class for algorithms that update weights.

    Warns
    -----
    It works with any algorithm based on the backpropagation.
    """
    addon_type = WEIGHT_PENALTY



import tensorflow as tf

from neupy.core.properties import NumberProperty
from .base import MinibatchGradientDescent


__all__ = ('Adagrad',)


class Adagrad(MinibatchGradientDescent):
    """
    Adagrad algorithm.

    Parameters
    ----------
    epsilon : float
        Value need to be greater than ``0``.
        Defaults to ``1e-5``.

    {MinibatchGradientDescent.Parameters}

    Attributes
    ----------
    {MinibatchGradientDescent.Attributes}

    Methods
    -------
    {MinibatchGradientDescent.Methods}

    Examples
    --------
    >>> import numpy as np
    >>> from neupy import algorithms
    >>>
    >>> x_train = np.array([[1, 2], [3, 4]])
    >>> y_train = np.array([[1], [0]])
    >>>
    >>> mnet = algorithms.Adagrad((2, 3, 1))
    >>> mnet.train(x_train, y_train)

    References
    ----------
    [1] John Duchi, Elad Hazan, Yoram Singer,
        Adaptive Subgradient Methods for Online Learning and Stochastic
        Optimization
        http://www.jmlr.org/papers/volume12/duchi11a/duchi11a.pdf
    """
    epsilon = NumberProperty(default=1e-5, minval=0)

    def init_train_updates(self):
        updates = []
        step = self.variables.step

        for layer, parameter, gradient in self.iter_params_and_grads():
            prev_mean_squred_grad = tf.Variable(
                tf.zeros(parameter.shape),
                name="{}/prev-mean-squred-grad".format(parameter.op.name),
                dtype=tf.float32,
            )

            mean_squred_grad = prev_mean_squred_grad + gradient ** 2
            parameter_delta = gradient / (
                tf.sqrt(mean_squred_grad + self.epsilon))

            updates.extend([
                (prev_mean_squred_grad, mean_squred_grad),
                (parameter, parameter - step * parameter_delta),
            ])

        return updates



import numpy as np
import tensorflow as tf

from neupy import init
from neupy.utils import asfloat, as_tuple
from neupy.core.properties import (NumberProperty, TypedListProperty,
                                   ParameterProperty, IntProperty)
from .utils import dimshuffle
from .base import ParameterBasedLayer


__all__ = ('ActivationLayer', 'Linear', 'Sigmoid', 'HardSigmoid', 'Tanh',
           'Relu', 'Softplus', 'Softmax', 'Elu', 'PRelu', 'LeakyRelu')


class ActivationLayer(ParameterBasedLayer):
    """
    Base class for the layers based on the activation
    functions.

    Parameters
    ----------
    size : int or None
        Layer input size. ``None`` means that layer will not create
        parameters and will return only activation function
        output for the specified input value.

    {ParameterBasedLayer.weight}

    {ParameterBasedLayer.bias}

    {BaseLayer.Parameters}

    Methods
    -------
    {ParameterBasedLayer.Methods}

    Attributes
    ----------
    {ParameterBasedLayer.Attributes}
    """
    size = IntProperty(minval=1, default=None, allow_none=True)

    def __init__(self, size=None, **options):
        super(ActivationLayer, self).__init__(size=size, **options)

    @property
    def output_shape(self):
        if self.size is not None:
            return as_tuple(self.size)
        return self.input_shape

    def initialize(self):
        if self.size is not None:
            super(ActivationLayer, self).initialize()

    def output(self, input_value):
        if self.size is not None:
            input_value = tf.matmul(input_value, self.weight)

            if self.bias is not None:
                input_value += self.bias

        return self.activation_function(input_value)

    def __repr__(self):
        if self.size is None:
            return super(ParameterBasedLayer, self).__repr__()
        return super(ActivationLayer, self).__repr__()


class Linear(ActivationLayer):
    """
    The layer with the linear activation function.

    Parameters
    ----------
    {ActivationLayer.Parameters}

    Methods
    -------
    {ActivationLayer.Methods}

    Attributes
    ----------
    {ActivationLayer.Attributes}
    """
    def activation_function(self, input_value):
        return input_value


class Sigmoid(ActivationLayer):
    """
    The layer with the sigmoid activation function.

    Parameters
    ----------
    {ActivationLayer.Parameters}

    Methods
    -------
    {ActivationLayer.Methods}

    Attributes
    ----------
    {ActivationLayer.Attributes}
    """
    def activation_function(self, input_value):
        return tf.nn.sigmoid(input_value)


class HardSigmoid(ActivationLayer):
    """
    The layer with the hard sigmoid activation function.

    Parameters
    ----------
    {ActivationLayer.Parameters}

    Methods
    -------
    {ActivationLayer.Methods}

    Attributes
    ----------
    {ActivationLayer.Attributes}
    """
    def activation_function(self, input_value):
        input_value = (0.2 * input_value) + 0.5
        return tf.clip_by_value(input_value, 0., 1.)


class Tanh(ActivationLayer):
    """
    The layer with the `tanh` activation function.

    Parameters
    ----------
    {ActivationLayer.Parameters}

    Methods
    -------
    {ActivationLayer.Methods}

    Attributes
    ----------
    {ActivationLayer.Attributes}
    """
    def activation_function(self, input_value):
        return tf.nn.tanh(input_value)


class Relu(ActivationLayer):
    """
    The layer with the rectifier (ReLu) activation function.

    Parameters
    ----------
    alpha : float
        Alpha parameter defines the decreasing rate
        for the negative values. If ``alpha``
        is non-zero value then layer behave like a
        leaky ReLu. Defaults to ``0``.

    {ActivationLayer.Parameters}

    Methods
    -------
    {ActivationLayer.Methods}

    Attributes
    ----------
    {ActivationLayer.Attributes}
    """
    alpha = NumberProperty(default=0, minval=0)
    weight = ParameterProperty(default=init.XavierNormal(gain='relu'))

    def activation_function(self, input_value):
        if self.alpha == 0:
            return tf.nn.relu(input_value)
        return tf.nn.leaky_relu(input_value, asfloat(self.alpha))


class LeakyRelu(ActivationLayer):
    """
    The layer with the leaky rectifier (Leaky ReLu)
    activation function.

    Parameters
    ----------
    {ActivationLayer.Parameters}

    Methods
    -------
    {ActivationLayer.Methods}

    Attributes
    ----------
    {ActivationLayer.Attributes}

    Notes
    -----
    Do the same as ``layers.Relu(input_size, alpha=0.01)``.
    """
    def activation_function(self, input_value):
        return tf.nn.leaky_relu(input_value, alpha=asfloat(0.01))


class Softplus(ActivationLayer):
    """
    The layer with the softplus activation function.

    Parameters
    ----------
    {ActivationLayer.Parameters}

    Methods
    -------
    {ActivationLayer.Methods}

    Attributes
    ----------
    {ActivationLayer.Attributes}
    """
    def activation_function(self, input_value):
        return tf.nn.softplus(input_value)


class Softmax(ActivationLayer):
    """
    The layer with the softmax activation function.

    Parameters
    ----------
    {ActivationLayer.Parameters}

    Methods
    -------
    {ActivationLayer.Methods}

    Attributes
    ----------
    {ActivationLayer.Attributes}
    """
    def activation_function(self, input_value):
        return tf.nn.softmax(input_value)


class Elu(ActivationLayer):
    """
    The layer with the exponensial linear unit (ELU)
    activation function.

    Parameters
    ----------
    {ActivationLayer.Parameters}

    Methods
    -------
    {ActivationLayer.Methods}

    Attributes
    ----------
    {ActivationLayer.Attributes}

    References
    ----------
    .. [1] http://arxiv.org/pdf/1511.07289v3.pdf
    """
    def activation_function(self, input_value):
        return tf.nn.elu(input_value)


class AxesProperty(TypedListProperty):
    """
    Property defines axes parameter.

    Parameters
    ----------
    {TypedListProperty.n_elements}

    {TypedListProperty.element_type}

    {BaseProperty.default}

    {BaseProperty.required}
    """
    def __set__(self, instance, value):
        value = (value,) if isinstance(value, int) else value
        super(AxesProperty, self).__set__(instance, value)


class PRelu(ActivationLayer):
    """
    The layer with the parametrized ReLu activation
    function.

    Parameters
    ----------
    alpha_axes : int or tuple
        Axes that will not include unique alpha parameter.
        Single integer value defines the same as a tuple with one value.
        Defaults to ``-1``.

    alpha : array-like, Tensorfow variable, scalar or Initializer
        Alpha parameter per each non-shared axis for the ReLu.
        Scalar value means that each element in the tensor will be
        equal to the specified value.
        Default initialization methods you can find
        :ref:`here <init-methods>`.
        Defaults to ``Constant(value=0.25)``.

    {ActivationLayer.Parameters}

    Methods
    -------
    {ActivationLayer.Methods}

    Attributes
    ----------
    {ActivationLayer.Attributes}

    References
    ----------
    .. [1] https://arxiv.org/pdf/1502.01852v1.pdf
    """
    alpha_axes = AxesProperty(default=-1)
    alpha = ParameterProperty(default=init.Constant(value=0.25))

    def __init__(self, *args, **options):
        super(PRelu, self).__init__(*args, **options)

        if 0 in self.alpha_axes:
            raise ValueError("Cannot specify alpha for 0-axis")

    def validate(self, input_shape):
        if max(self.alpha_axes) > len(input_shape):
            max_axis_index = len(input_shape) - 1
            raise ValueError("Cannot specify alpha for the axis #{}. "
                             "Maximum available axis is {} (0-based indeces)."
                             "".format(max(self.alpha_axes), max_axis_index))

    def initialize(self):
        super(PRelu, self).initialize()
        output_shape = as_tuple(None, self.output_shape)

        alpha_shape = [output_shape[axis] for axis in self.alpha_axes]
        self.add_parameter(
            value=self.alpha,
            name='alpha',
            shape=alpha_shape,
            trainable=True,
        )

    def activation_function(self, input_value):
        input_value = tf.convert_to_tensor(input_value, dtype=tf.float32)
        ndim = len(input_value.get_shape())

        dimensions = np.arange(ndim)
        alpha_axes = dimensions[list(self.alpha_axes)]

        alpha = dimshuffle(self.alpha, ndim, alpha_axes)
        return tf.nn.leaky_relu(tf.to_float(input_value), tf.to_float(alpha))



import copy
import pprint
import inspect
from collections import OrderedDict

import six

from neupy.exceptions import LayerConnectionError


__all__ = ('LayerGraph',)


def filter_list(iterable, include_values):
    """
    Create new list that contains only values
    specified in the ``include_values`` attribute.

    Parameters
    ----------
    iterable : list
        List that needs to be filtered.

    include_values : list, tuple
        List of values that needs to be included in the
        filtered list. Other values that hasn't been
        defined in the list will be excluded from the
        list specified by ``iterable`` attribute.

    Returns
    -------
    list
        Filtered list.
    """
    filtered_list = []

    for value in iterable:
        if value in include_values:
            filtered_list.append(value)

    return filtered_list


def filter_dict(dictionary, include_keys):
    """
    Create new list that contains only values
    specified in the ``include_keys`` attribute.

    Parameters
    ----------
    dictionary : dict
        Original dictionary

    include_keys : list or tuple
        Keys that will copied from original dictionary
        into a new one.

    Returns
    -------
    dict
    """
    filtered_dict = OrderedDict()

    for key, value in dictionary.items():
        if key in include_keys:
            filtered_dict[key] = filter_list(value, include_keys)

    return filtered_dict


def merge_dicts_with_list(first_dict, second_dict):
    """
    Create new dict that contains all elements from the first
    one and from the second one. Function assumes that all value
    elements are lists. In case if one key appears in both dicts
    function will merge these lists into one.

    Parameters
    ----------
    first_dict : dict
    second_dict : dict

    Returns
    -------
    dict
    """
    common_dict = OrderedDict()

    for key, value in first_dict.items():
        # To make sure that we copied lists inside of the
        # dictionary, but didn't copied values inside of the list
        common_dict[key] = copy.copy(value)

    for key, values in second_dict.items():
        if key in common_dict:
            for value in values:
                if value not in common_dict[key]:
                    common_dict[key].append(value)
        else:
            common_dict[key] = copy.copy(values)

    return common_dict


def is_cyclic(graph):
    """
    Check if graph has cycles.

    Parameters
    ----------
    graph : dict
        must be represented as a dictionary mapping vertices to
        iterables of neighbouring vertices.

    Returns
    -------
    bool
        Return ``True`` if the directed graph has a cycle.

    Examples
    --------
    >>> is_cyclic({1: (2,), 2: (3,), 3: (1,)})
    True
    >>> is_cyclic({1: (2,), 2: (3,), 3: (4,)})
    False
    """
    path = set()
    visited = set()

    def visit(vertex):
        if vertex in visited:
            return False

        visited.add(vertex)
        path.add(vertex)

        for neighbour in graph.get(vertex, ()):
            if neighbour in path or visit(neighbour):
                return True

        path.remove(vertex)
        return False

    return any(visit(vertex) for vertex in graph)


def does_layer_expect_one_input(layer):
    """
    Check whether layer can except only one input layer.

    Parameters
    ----------
    layer : layer or connection

    Raises
    ------
    ValueError
        In case if argument is not a layer.

    Retruns
    -------
    bool
        Returns ``True`` if layer can accept onl one input
        layer, ``False`` otherwise.
    """
    if not hasattr(layer, 'output'):
        raise ValueError("Layer `{}` doesn't have "
                         "output method".format(layer))

    if not inspect.ismethod(layer.output):
        raise ValueError("Layer has an `output` property, "
                         "but it's not a method")

    # The main output method overwrapped with decoretor that destroys
    # original properties of the method. The original method can be found
    # in the `original_method` attribute
    arginfo = inspect.getargspec(layer.output.original_method)

    if arginfo.varargs is not None:
        return False

    # In case if layer expects fixed number of input layers
    n_args = len(arginfo.args) - 1  # Ignore `self` argument
    return n_args == 1


class LayerGraph(object):
    """
    Direct Acyclic Graph (DAG) for layer connections.

    Parameters
    ----------
    forward_graph : None or dict
    backward_graph : None or dict

    Raises
    ------
    LayerConnectionError
        If graph cannot connect layers.
    """
    def __init__(self, forward_graph=None, backward_graph=None):
        if forward_graph is None:
            forward_graph = OrderedDict()

        if backward_graph is None:
            backward_graph = OrderedDict()

        self.forward_graph = forward_graph
        self.backward_graph = backward_graph

    @classmethod
    def merge(cls, left_graph, right_graph):
        """
        Combine two separated graphs into one.

        Parameters
        ----------
        left_graph : LayerGraph instance
        right_graph : LayerGraph instance

        Returns
        -------
        LayerGraph instance
            New graph that contains layers and connections
            from input graphs.
        """
        forward_graph = merge_dicts_with_list(
            left_graph.forward_graph,
            right_graph.forward_graph
        )
        backward_graph = merge_dicts_with_list(
            left_graph.backward_graph,
            right_graph.backward_graph
        )
        return cls(forward_graph, backward_graph)

    def add_layer(self, layer):
        """
        Add new layer into the graph.

        Parameters
        ----------
        layer : layer

        Returns
        -------
        bool
            Returns ``False`` if layer has beed already added into
            graph and there is no need to add it again, and
            ``True`` - if layer is a new and was added successfully.
        """
        if layer in self.forward_graph:
            return False

        for existed_layer in self.forward_graph:
            if existed_layer.name == layer.name:
                raise LayerConnectionError(
                    "Cannot connect {} layer. Layer with name {!r} has been "
                    "already defined in the graph.".format(layer, layer.name))

        self.forward_graph[layer] = []
        self.backward_graph[layer] = []

        return True

    def add_connection(self, from_layer, to_layer):
        """
        Add new directional connection between two layers.

        Parameters
        ----------
        from_layer : layer
        to_layer : layer

        Raises
        ------
        LayerConnectionError
            Raises if it's impossible to connect two layers or
            new connection creates cycles in graph.

        Returns
        -------
        bool
            Returns ``False`` if connection has already been added into
            the graph, and ``True`` if connection was added successfully.
        """
        if from_layer is to_layer:
            raise LayerConnectionError("Cannot connect layer `{}` "
                                       "to itself".format(from_layer))

        self.add_layer(from_layer)
        self.add_layer(to_layer)

        forward_connections = self.forward_graph[from_layer]
        backward_connections = self.backward_graph[to_layer]

        if to_layer in forward_connections:
            # Layers have been already connected
            return False

        forward_connections.append(to_layer)
        backward_connections.append(from_layer)

        if is_cyclic(self.forward_graph):
            raise LayerConnectionError(
                "Cannot connect layer `{}` to `{}`, because this "
                "connection creates cycle in the graph."
                "".format(from_layer, to_layer))

        return True

    def connect_layers(self, from_layers, to_layers):
        """
        Connect two layers together and update other layers
        in the graph.

        Parameters
        ----------
        from_layer : layer or list of layers
        to_layer : layer or list of layers

        Raises
        ------
        LayerConnectionError
            Raises if cannot graph cannot connect two layers.

        Returns
        -------
        bool
            Returns ``False`` if connection has already been added into
            the graph, and ``True`` if connection was added successfully.
        """
        if not isinstance(from_layers, (list, tuple)):
            from_layers = [from_layers]

        if not isinstance(to_layers, (list, tuple)):
            to_layers = [to_layers]

        connections_added = []
        do_not_have_shapes = True

        for from_layer in from_layers:
            if from_layer.input_shape or from_layer.output_shape:
                do_not_have_shapes = False

            for to_layer in to_layers:
                connection_added = self.add_connection(from_layer, to_layer)
                connections_added.append(connection_added)

        if not any(connections_added):
            return False

        if do_not_have_shapes:
            return True

        # Layer has an input shape which means that we can
        # propagate this information through the graph and
        # set up input shape for layers that don't have it.
        layers = copy.copy(from_layers)
        forward_graph = self.forward_graph

        # We need to know whether all input layers
        # have defined input shape
        all_inputs_has_shape = all(
            layer.input_shape for layer in self.input_layers)

        while layers:
            current_layer = layers.pop()
            next_layers = forward_graph[current_layer]

            for next_layer in next_layers:
                next_inp_shape = next_layer.input_shape
                current_out_shape = current_layer.output_shape
                expect_one_input = does_layer_expect_one_input(next_layer)

                if not next_inp_shape and expect_one_input:
                    next_layer.input_shape = current_out_shape
                    next_layer.initialize()

                elif not expect_one_input and all_inputs_has_shape:
                    input_shapes = []
                    for incoming_layer in self.backward_graph[next_layer]:
                        input_shapes.append(incoming_layer.output_shape)

                    if None not in input_shapes:
                        next_layer.input_shape = input_shapes
                        next_layer.initialize()

                    else:
                        # Some of the previous layers still don't
                        # have input shape. We can put layer at the
                        # end of the stack and check it again at the end
                        layers.insert(0, current_layer)

                elif expect_one_input and next_inp_shape != current_out_shape:
                    raise LayerConnectionError(
                        "Cannot connect `{}` to the `{}`. Output shape "
                        "from one layer is equal to {} and input shape "
                        "to the next one is equal to {}".format(
                            current_layer, next_layer,
                            current_out_shape, next_inp_shape,
                        ))

            layers.extend(next_layers)

        return True

    def reverse(self):
        """
        Returns graph with reversed connections.
        """
        return LayerGraph(self.backward_graph, self.forward_graph)

    def subgraph_for_output(self, output_layers):
        """
        Extract subgraph with specified set
        of output layers.

        Parameters
        ----------
        layers : layer, list of layers

        Returns
        -------
        LayerGraph instance
        """
        if not isinstance(output_layers, (list, tuple)):
            output_layers = [output_layers]

        if all(layer not in self.forward_graph for layer in output_layers):
            return LayerGraph()

        observed_layers = []
        layers = copy.copy(output_layers)

        while layers:
            current_layer = layers.pop()
            next_layers = self.backward_graph[current_layer]

            for next_layer in next_layers:
                if next_layer not in observed_layers:
                    layers.append(next_layer)

            observed_layers.append(current_layer)

        forward_subgraph = filter_dict(self.forward_graph, observed_layers)
        backward_subgraph = filter_dict(self.backward_graph, observed_layers)

        # Remove old relations to the other layers.
        # Output layer cannot point to some other layers.
        for layer in output_layers:
            forward_subgraph[layer] = []

        return LayerGraph(forward_subgraph, backward_subgraph)

    def subgraph_for_input(self, input_layers):
        """
        Extract subgraph with specified set
        of input layers.

        Parameters
        ----------
        layers : layer, list of layers

        Returns
        -------
        LayerGraph instance
        """
        # Output layers for the reversed graph are
        # input layers for normal graph
        graph_reversed = self.reverse()
        subgraph_reversed = graph_reversed.subgraph_for_output(input_layers)

        # Reverse it to make normal graph
        return subgraph_reversed.reverse()

    def subgraph(self, input_layers, output_layers):
        """
        Create subgraph that contains only layers that
        has relations between specified input and output
        layers.

        Parameters
        ----------
        input_layers : layer, list of layers
        output_layers : layer, list of layers

        Returns
        -------
        LayerGraph
        """
        subgraph = self.subgraph_for_input(input_layers)
        return subgraph.subgraph_for_output(output_layers)

    @property
    def input_layers(self):
        """
        List of input layers.

        Raises
        ------
        LayerConnectionError
            If graph doesn't have input layers.

        Returns
        -------
        list
            List of input layers.
        """
        input_layers = []

        for layer, next_layers in self.backward_graph.items():
            if not next_layers:
                input_layers.append(layer)

        return input_layers

    @property
    def output_layers(self):
        """
        List of output layers.

        Raises
        ------
        LayerConnectionError
            If graph doesn't have output layers.

        Returns
        -------
        list
            List of output layers.
        """
        output_layers = []

        for layer, next_layers in self.forward_graph.items():
            if not next_layers:
                output_layers.append(layer)

        return output_layers

    def find_layer_by_name(self, layer_name):
        """
        Find layer instance in the graph based on the
        specified layer name.

        Parameters
        ----------
        layer_name : str
            Name of the layer that presented in this graph.

        Raises
        ------
        NameError
            In case if there is no layer with specified
            name in the graph.

        Returns
        -------
        layer
        """
        for layer in self.forward_graph:
            if layer.name == layer_name:
                return layer

        raise NameError("Cannot find layer with name {!r}".format(layer_name))

    def propagate_forward(self, input_value):
        """
        Propagates input variable through the directed acyclic
        graph and returns output from the final layers.

        Parameters
        ----------
        input_value : array-like, Tensorfow variable or dict
            - If input is an array or Tensorfow variable than it will
              be used as a direct input to the input layer/layers.

            - The dict type input should has a specific structure.
              Each key of the dict is a layer and each value array or
              Tensorfow variable. Dict defines input values for specific
              layers. In the dict input layer is not necessary should
              be an instance of the ``layers.Input`` class. It can be
              any layer from the graph.

        Returns
        -------
        object
            Output from the final layer/layers.
        """
        outputs = {}

        if isinstance(input_value, dict):
            for layer, input_variable in input_value.items():
                if isinstance(layer, six.string_types):
                    layer = self.find_layer_by_name(layer)

                if layer not in self.forward_graph:
                    raise ValueError("The `{}` layer doesn't appear "
                                     "in the graph".format(layer))

                outputs[layer] = layer.output(input_variable)
        else:
            for input_layer in self.input_layers:
                outputs[input_layer] = input_layer.output(input_value)

        def output_from_layer(layer):
            if layer in outputs:
                return outputs[layer]

            input_layers = self.backward_graph[layer]
            inputs = []
            for input_layer in input_layers:
                if input_layer in outputs:
                    res = outputs[input_layer]
                else:
                    res = output_from_layer(input_layer)
                    outputs[input_layer] = res

                inputs.append(res)

            return layer.output(*inputs)

        results = []
        for output_layer in self.output_layers:
            results.append(output_from_layer(output_layer))

        if len(results) == 1:
            results = results[0]

        return results

    def layer_names_only(self):
        """
        Replaces in the graph layers with their names.

        Parameters
        ----------
        graph : LayerGraph

        Returns
        -------
        OrderedDict
        """
        prepared_graph = OrderedDict()

        for from_layer, to_layers in self.forward_graph.items():
            prepared_graph[from_layer.name] = [l.name for l in to_layers]

        return list(prepared_graph.items())

    def __len__(self):
        return len(self.forward_graph)

    def __repr__(self):
        graph = list(self.forward_graph.items())
        return pprint.pformat(graph)



import os
import random

import tensorflow as tf
import numpy as np


__all__ = ('reproducible',)


def reproducible(seed=0):
    """
    Set up the same seed value for the NumPy and
    python random module to make your code reproducible.

    Parameters
    ----------
    seed : int
        Defaults to ``0``.
    """
    os.environ['PYTHONHASHSEED'] = str(seed)
    np.random.seed(seed)
    random.seed(seed)
    tf.set_random_seed(seed)



from __future__ import division

import math
import collections

import six
import tensorflow as tf

from neupy.utils import as_tuple
from neupy.exceptions import LayerConnectionError
from neupy.core.properties import (TypedListProperty, Property,
                                   WithdrawProperty)
from .base import ParameterBasedLayer


__all__ = ('Convolution', 'Deconvolution')


class Spatial2DProperty(TypedListProperty):
    """
    Stride property.

    Parameters
    ----------
    {BaseProperty.Parameters}
    """
    expected_type = (list, tuple, int)

    def __init__(self, *args, **kwargs):
        kwargs['element_type'] = int
        super(Spatial2DProperty, self).__init__(*args, **kwargs)

    def __set__(self, instance, value):
        if isinstance(value, collections.Iterable) and len(value) == 1:
            value = (value[0], 1)

        if isinstance(value, int):
            value = (value, value)

        super(Spatial2DProperty, self).__set__(instance, value)

    def validate(self, value):
        super(Spatial2DProperty, self).validate(value)

        if len(value) > 2:
            raise ValueError("Stide can have only one or two elements "
                             "in the list. Got {}".format(len(value)))

        if any(element <= 0 for element in value):
            raise ValueError(
                "Stride size should contain only values greater than zero")


def deconv_output_shape(dimension_size, filter_size, padding, stride,
                        diation=1):
    """
    Computes deconvolution's output shape for one spatial dimension.

    Parameters
    ----------
    dimension_size : int or None
        Size of the dimension. Typically it's image's weight or height.
        It might be equal to ``None`` when we input might have variable
        dimension.

    filter_size : int
        Size of the convolution filter.

    padding : {``valid``, ``same``} or int
        Type or size of the zero-padding.

    stride : int
        Stride size.

    Returns
    -------
    int
        Dimension size after applying deconvolution
        operation with specified configurations.
    """
    if dimension_size is None:
        return None

    if padding in ('VALID', 'valid'):
        return dimension_size * stride + max(filter_size - stride, 0)

    elif padding in ('SAME', 'same'):
        return dimension_size * stride

    elif isinstance(padding, int):
        return dimension_size * stride - 2 * padding + filter_size - 1

    raise ValueError("`{!r}` is unknown deconvolution's padding value"
                     "".format(padding))


def conv_output_shape(dimension_size, filter_size, padding, stride,
                      dilation=1):
    """
    Computes convolution's output shape for one spatial dimension.

    Parameters
    ----------
    dimension_size : int or None
        Size of the dimension. Typically it's image's weight or height.
        It might be equal to ``None`` when we input might have variable
        dimension.

    filter_size : int
        Size of the convolution filter.

    padding : {``valid``, ``same``} or int
        Type or size of the zero-padding.

    stride : int
        Stride size.

    dilation : int
        Dilation rate.

    Returns
    -------
    int
        Dimension size after applying convolution
        operation with specified configurations.
    """
    if dimension_size is None:
        return None

    if not isinstance(stride, int):
        raise ValueError("Stride needs to be an integer, got {} (value {!r})"
                         "".format(type(stride), stride))

    if not isinstance(filter_size, int):
        raise ValueError("Filter size needs to be an integer, got {} "
                         "(value {!r})".format(type(filter_size),
                                               filter_size))

    # We can think of the dilation as very sparse convolutional fitler
    # filter=3 and diation=2 the same as filter=5 and dilation=1
    filter_size = filter_size + (filter_size - 1) * (dilation - 1)

    if padding in ('VALID', 'valid'):
        return int(math.ceil((dimension_size - filter_size + 1) / stride))

    elif padding in ('SAME', 'same'):
        return int(math.ceil(dimension_size / stride))

    elif isinstance(padding, int):
        return int(math.ceil(
            (dimension_size + 2 * padding - filter_size + 1) / stride))

    raise ValueError("`{!r}` is unknown convolution's padding value"
                     "".format(padding))


class PaddingProperty(Property):
    """
    Border mode property identifies border for the
    convolution operation.
    Parameters
    ----------
    {Property.Parameters}
    """
    expected_type = (six.string_types, int, tuple)
    valid_string_choices = ('VALID', 'SAME', 'same', 'valid')

    def __set__(self, instance, value):
        if isinstance(value, int):
            if value < 0:
                raise ValueError(
                    "Integer border mode value needs to be "
                    "greater or equal to zero, got {}".format(value)
                )

            value = (value, value)

        if isinstance(value, six.string_types):
            value = value.upper()

        super(PaddingProperty, self).__set__(instance, value)

    def validate(self, value):
        super(PaddingProperty, self).validate(value)

        if isinstance(value, tuple) and len(value) != 2:
            raise ValueError(
                "Border mode property suppose to get a tuple that "
                "contains two elements, got {} elements"
                "".format(len(value))
            )

        is_invalid_string = (
            isinstance(value, six.string_types) and
            value not in self.valid_string_choices
        )

        if is_invalid_string:
            valid_choices = ', '.join(self.valid_string_choices)
            raise ValueError("`{}` is invalid string value. Available: {}"
                             "".format(value, valid_choices))

        if isinstance(value, tuple) and any(element < 0 for element in value):
            raise ValueError("Tuple border mode value needs to contain "
                             "only elements that greater or equal to zero, "
                             "got {}".format(value))


class Convolution(ParameterBasedLayer):
    """
    Convolutional layer.

    Parameters
    ----------
    size : tuple of int
        Filter shape. In should be defined as a tuple with three
        integers ``(filter rows, filter columns, output channels)``.

    padding : {{``same``, ``valid``}}, int, tuple
        Zero padding for the input tensor.

        - ``valid`` - Padding won't be added to the tensor. Result will be
          the same as for ``padding=0``

        - ``same`` - Padding will depend on the number of rows and columns
          in the filter. This padding makes sure that image with the
          ``stride=1`` won't change its width and height. It's the same as
          ``padding=(filter rows // 2, filter columns // 2)``.

        - Custom value for the padding can be specified as an integer, like
          ``padding=1`` or it can be specified as a tuple when different
          dimensions have different padding values, for example
          ``padding=(2, 3)``.

        Defaults to ``valid``.

    stride : tuple with ints, int.
        Stride size. Defaults to ``(1, 1)``

    dilation : int, tuple
        Rate for the fiter upsampling. When ``dilation > 1`` layer will
        become diated convolution (or atrous convolution). Defaults to ``1``.

    weight : array-like, Tensorfow variable, scalar or Initializer
        Defines layer's weights. Shape of the weight will be equal to
        ``(filter rows, filter columns, input channels, output channels)``.
        Default initialization methods you can find
        :ref:`here <init-methods>`. Defaults to
        :class:`XavierNormal() <neupy.init.XavierNormal>`.

    {ParameterBasedLayer.bias}

    {BaseLayer.Parameters}

    Examples
    --------
    2D Convolution

    >>> from neupy import layers
    >>>
    >>> layers.join(
    ...     layers.Input((28, 28, 3)),
    ...     layers.Convolution((3, 3, 16)),
    ... )

    1D Convolution

    >>> from neupy import layers
    >>>
    >>> layers.join(
    ...     layers.Input((30, 10)),
    ...     layers.Reshape((30, 1, 10)),
    ...     layers.Convolution((3, 1, 16)),
    ... )

    Methods
    -------
    {ParameterBasedLayer.Methods}

    Attributes
    ----------
    {ParameterBasedLayer.Attributes}
    """
    size = TypedListProperty(required=True, element_type=int)
    padding = PaddingProperty(default='valid')
    stride = Spatial2DProperty(default=(1, 1))
    dilation = Spatial2DProperty(default=1)

    def validate(self, input_shape):
        if len(input_shape) != 3:
            raise LayerConnectionError(
                "Convolutional layer expects an input with 3 "
                "dimensions, got {} with shape {}"
                "".format(len(input_shape), input_shape))

    def output_shape_per_dim(self, *args, **kwargs):
        return conv_output_shape(*args, **kwargs)

    def find_output_from_input_shape(self, input_shape):
        padding = self.padding
        rows, cols, _ = input_shape

        row_filter_size, col_filter_size, n_kernels = self.size
        row_stride, col_stride = self.stride
        row_dilation, col_dilation = self.dilation or (1, 1)

        if isinstance(padding, (list, tuple)):
            row_padding, col_padding = padding
        else:
            row_padding, col_padding = padding, padding

        output_rows = self.output_shape_per_dim(
            rows, row_filter_size,
            row_padding, row_stride, row_dilation,
        )
        output_cols = self.output_shape_per_dim(
            cols, col_filter_size,
            col_padding, col_stride, col_dilation,
        )

        return (output_rows, output_cols, n_kernels)

    @property
    def output_shape(self):
        if self.input_shape is not None:
            return self.find_output_from_input_shape(self.input_shape)

    @property
    def weight_shape(self):
        n_channels = self.input_shape[-1]
        n_rows, n_cols, n_filters = self.size
        return (n_rows, n_cols, n_channels, n_filters)

    @property
    def bias_shape(self):
        return as_tuple(self.size[-1])

    def output(self, input_value):
        padding = self.padding

        if not isinstance(padding, six.string_types):
            height_pad, weight_pad = padding
            input_value = tf.pad(input_value, [
                [0, 0],
                [height_pad, height_pad],
                [weight_pad, weight_pad],
                [0, 0],
            ])
            # VALID option will make sure that
            # convolution won't use any padding.
            padding = 'VALID'

        output = tf.nn.convolution(
            input_value,
            self.weight,
            padding=padding,
            strides=self.stride,
            dilation_rate=self.dilation,
            data_format="NHWC"
        )

        if self.bias is not None:
            bias = tf.reshape(self.bias, (1, 1, 1, -1))
            output += bias

        return output


class Deconvolution(Convolution):
    """
    Deconvolution layer. It's commonly called like this in the literature,
    but it's just gradient of the convolution and not actual deconvolution.

    Parameters
    ----------
    {Convolution.size}

    {Convolution.padding}

    {Convolution.stride}

    weight : array-like, Tensorfow variable, scalar or Initializer
        Defines layer's weights. Shape of the weight will be equal to
        ``(filter rows, filter columns, output channels, input channels)``.
        Default initialization methods you can find
        :ref:`here <init-methods>`. Defaults to
        :class:`XavierNormal() <neupy.init.XavierNormal>`.

    {ParameterBasedLayer.bias}

    {BaseLayer.Parameters}

    Methods
    -------
    {ParameterBasedLayer.Methods}

    Examples
    --------
    >>> from neupy import layers
    >>>
    >>> layers.join(
    ...     layers.Input((28, 28, 3)),
    ...     layers.Convolution((3, 3, 16)),
    ...     layers.Deconvolution((3, 3, 1)),
    ... )

    Attributes
    ----------
    {ParameterBasedLayer.Attributes}
    """
    dilation = WithdrawProperty()

    def output_shape_per_dim(self, *args, **kwargs):
        return deconv_output_shape(*args, **kwargs)

    @property
    def weight_shape(self):
        # Compare to the regular convolution weights
        # have switched input and output channels.
        return as_tuple(self.size, self.input_shape[-1])

    def output(self, input_value):
        input_shape = tf.shape(input_value)
        # We need to get information about output shape from the input
        # tensor's shape, because for some inputs we might have
        # height and width specified as None and shape value won't be
        # computed for these dimensions.
        output_shape = self.find_output_from_input_shape(
            tf.unstack(input_shape[1:]))

        batch_size = input_shape[0]
        padding = self.padding

        if isinstance(self.padding, (list, tuple)):
            height_pad, width_pad = self.padding

            # VALID option will make sure that
            # deconvolution won't use any padding.
            padding = 'VALID'

            # conv2d_transpose doesn't know about extra paddings that we added
            # in the convolution. For this reason we have to expand our
            # expected output shape and later we will remove these paddings
            # manually after transpose convolution.
            output_shape = (
                output_shape[0] + 2 * height_pad,
                output_shape[1] + 2 * width_pad,
                output_shape[2],
            )

        output = tf.nn.conv2d_transpose(
            input_value,
            self.weight,
            as_tuple(batch_size, output_shape),
            as_tuple(1, self.stride, 1),
            padding,
            data_format="NHWC"
        )

        if isinstance(self.padding, (list, tuple)):
            h_pad, w_pad = self.padding

            if h_pad > 0:
                output = output[:, h_pad:-h_pad, :, :]

            if w_pad > 0:
                output = output[:, :, w_pad:-w_pad, :]

        if self.bias is not None:
            bias = tf.reshape(self.bias, (1, 1, 1, -1))
            output += bias

        return output



from __future__ import division

import numpy as np

from neupy.utils import format_data
from neupy.core.properties import (ProperFractionProperty,
                                   IntProperty)
from neupy.algorithms.base import BaseNetwork


__all__ = ('ART1',)


class ART1(BaseNetwork):
    """
    Adaptive Resonance Theory (ART1) Network for binary
    data clustering.

    Notes
    -----
    - Weights are not random, so the result will be
      always reproduceble.

    Parameters
    ----------
    rho : float
        Control reset action in training process. Value must be
        between ``0`` and ``1``, defaults to ``0.5``.

    n_clusters : int
        Number of clusters, defaults to ``2``. Min value is also ``2``.

    {BaseNetwork.step}

    {BaseNetwork.show_epoch}

    {BaseNetwork.shuffle_data}

    {BaseNetwork.epoch_end_signal}

    {BaseNetwork.train_end_signal}

    Methods
    -------
    train(input_data)
        ART trains until all clusters are found.

    predict(input_data)
        Each prediction trains a new network. Basically it's
        an alias to the ``train`` method.

    {BaseSkeleton.fit}

    Examples
    --------
    >>> import numpy as np
    >>> from neupy import algorithms
    >>>
    >>> data = np.array([
    ...     [0, 1, 0],
    ...     [1, 0, 0],
    ...     [1, 1, 0],
    ... ])
    >>>>
    >>> artnet = algorithms.ART1(
    ...     step=2,
    ...     rho=0.7,
    ...     n_clusters=2,
    ...     verbose=False
    ... )
    >>> artnet.predict(data)
    array([ 0.,  1.,  1.])
    """
    rho = ProperFractionProperty(default=0.5)
    n_clusters = IntProperty(default=2, minval=2)

    def train(self, input_data):
        input_data = format_data(input_data)

        if input_data.ndim != 2:
            raise ValueError("Input value must be 2 dimensional, got "
                             "{}".format(input_data.ndim))

        n_samples, n_features = input_data.shape
        n_clusters = self.n_clusters
        step = self.step
        rho = self.rho

        if np.any((input_data != 0) & (input_data != 1)):
            raise ValueError("ART1 Network works only with binary matrices")

        if not hasattr(self, 'weight_21'):
            self.weight_21 = np.ones((n_features, n_clusters))

        if not hasattr(self, 'weight_12'):
            scaler = step / (step + n_clusters - 1)
            self.weight_12 = scaler * self.weight_21.T

        weight_21 = self.weight_21
        weight_12 = self.weight_12

        if n_features != weight_21.shape[0]:
            raise ValueError("Input data has invalid number of features. "
                             "Got {} instead of {}"
                             "".format(n_features, weight_21.shape[0]))

        classes = np.zeros(n_samples)

        # Train network
        for i, p in enumerate(input_data):
            disabled_neurons = []
            reseted_values = []
            reset = True

            while reset:
                output1 = p
                input2 = np.dot(weight_12, output1.T)

                output2 = np.zeros(input2.size)
                input2[disabled_neurons] = -np.inf
                winner_index = input2.argmax()
                output2[winner_index] = 1

                expectation = np.dot(weight_21, output2)
                output1 = np.logical_and(p, expectation).astype(int)

                reset_value = np.dot(output1.T, output1) / np.dot(p.T, p)
                reset = reset_value < rho

                if reset:
                    disabled_neurons.append(winner_index)
                    reseted_values.append((reset_value, winner_index))

                if len(disabled_neurons) >= n_clusters:
                    # Got this case only if we test all possible clusters
                    reset = False
                    winner_index = None

                if not reset:
                    if winner_index is not None:
                        weight_12[winner_index, :] = (step * output1) / (
                            step + np.dot(output1.T, output1) - 1
                        )
                        weight_21[:, winner_index] = output1
                    else:
                        # Get result with the best `rho`
                        winner_index = max(reseted_values)[1]

                    classes[i] = winner_index

        return classes

    def predict(self, input_data):
        return self.train(input_data)



import numpy as np

from neupy.core.base import BaseSkeleton
from neupy.core.properties import ChoiceProperty, IntProperty
from neupy.core.config import Configurable


__all__ = ('DiscreteMemory',)


class DiscreteMemory(BaseSkeleton, Configurable):
    """
    Base class for discrete memory networks.

    Notes
    -----
    - Input and output vectors should contain only binary values.

    Parameters
    ----------
    mode : {{``sync``, ``async``}}
        Indentify pattern recovery mode.

        - ``sync`` mode tries to recover pattern using all
          values from the input vector.

        - ``async`` mode choose randomly some values from the
          input vector and iteratively repeat this procedure.
          Number of iterations defines by the ``n_times``
          parameter.

        Defaults to ``sync``.

    n_times : int
        Available only in ``async`` mode. Identify number
        of random trials. Defaults to ``100``.
    """

    mode = ChoiceProperty(default='sync', choices=['async', 'sync'])
    n_times = IntProperty(default=100, minval=1)

    def __init__(self, **options):
        super(DiscreteMemory, self).__init__(**options)
        self.weight = None

        if 'n_times' in options and self.mode != 'async':
            self.logs.warning("You can use `n_times` property only in "
                              "`async` mode.")

    def discrete_validation(self, matrix):
        """
        Validate discrete matrix.

        Parameters
        ----------
        matrix : array-like
            Matrix for validation.
        """
        if np.any(~np.isin(matrix, [0, 1])):
            raise ValueError(
                "This network expects only descrete inputs. It mean that "
                "it's possible to can use only matrices with binary values "
                "(0 and 1)."
            )



from functools import partial

from neupy import layers


__all__ = ('vgg16',)


def vgg16():
    """
    VGG16 network architecture with random parameters. Parameters
    can be loaded using ``neupy.storage`` module.

    Originally VGG16 was built in order to solve image classification
    problem. It was used in the ImageNet competition. The goal of the
    competition is to build a model that classifies image into one of
    the 1,000 categories. Categories include animals, objects, transports
    and so on.

    VGG16 has roughly 138 million parameters.

    Examples
    --------
    >>> from neupy import architectures
    >>> vgg16 = architectures.vgg16()
    >>> vgg16
    (224, 224, 3) -> [... 38 layers ...] -> 1000
    >>>
    >>> from neupy import algorithms
    >>> network = algorithms.Momentum(vgg16, verbose=True)

    See Also
    --------
    :architecture:`vgg19` : VGG19 network
    :architecture:`squeezenet` : SqueezeNet network
    :architecture:`resnet50` : ResNet50 network

    References
    ----------
    Very Deep Convolutional Networks for Large-Scale Image Recognition.
    https://arxiv.org/abs/1409.1556
    """
    SamePadConvolution = partial(layers.Convolution, padding='SAME')

    return layers.join(
        layers.Input((224, 224, 3)),

        SamePadConvolution((3, 3, 64), name='conv1_1') > layers.Relu(),
        SamePadConvolution((3, 3, 64), name='conv1_2') > layers.Relu(),
        layers.MaxPooling((2, 2)),

        SamePadConvolution((3, 3, 128), name='conv2_1') > layers.Relu(),
        SamePadConvolution((3, 3, 128), name='conv2_2') > layers.Relu(),
        layers.MaxPooling((2, 2)),

        SamePadConvolution((3, 3, 256), name='conv3_1') > layers.Relu(),
        SamePadConvolution((3, 3, 256), name='conv3_2') > layers.Relu(),
        SamePadConvolution((3, 3, 256), name='conv3_3') > layers.Relu(),
        layers.MaxPooling((2, 2)),

        SamePadConvolution((3, 3, 512), name='conv4_1') > layers.Relu(),
        SamePadConvolution((3, 3, 512), name='conv4_2') > layers.Relu(),
        SamePadConvolution((3, 3, 512), name='conv4_3') > layers.Relu(),
        layers.MaxPooling((2, 2)),

        SamePadConvolution((3, 3, 512), name='conv5_1') > layers.Relu(),
        SamePadConvolution((3, 3, 512), name='conv5_2') > layers.Relu(),
        SamePadConvolution((3, 3, 512), name='conv5_3') > layers.Relu(),
        layers.MaxPooling((2, 2)),

        layers.Reshape(),

        layers.Linear(4096, name='dense_1') > layers.Relu(),
        layers.Dropout(0.5),

        layers.Linear(4096, name='dense_2') > layers.Relu(),
        layers.Dropout(0.5),

        layers.Linear(1000, name='dense_3') > layers.Softmax(),
    )



from neupy.core.docs import SharedDocs
from neupy.core.properties import WithdrawProperty


__all__ = ('LazyLearningMixin',)


class LazyLearningMixin(SharedDocs):
    """
    Mixin for lazy learning Neural Network algorithms.

    Notes
    -----
    - Network uses lazy learning which mean that network doesn't
      need iterative training. It just stores parameters
      and use them to make a predictions.

    Methods
    -------
    train(input_train, target_train, copy=True)
        Network just stores all the information about the data and use
        it for the prediction. Parameter ``copy`` copies input data
        before saving it inside the network.
    """
    step = WithdrawProperty()
    show_epoch = WithdrawProperty()
    shuffle_data = WithdrawProperty()
    train_end_signal = WithdrawProperty()
    epoch_end_signal = WithdrawProperty()

    def __init__(self, *args, **kwargs):
        self.input_train = None
        self.target_train = None
        super(LazyLearningMixin, self).__init__(*args, **kwargs)

    def train(self, input_train, target_train):
        self.input_train = input_train
        self.target_train = target_train

        if input_train.shape[0] != target_train.shape[0]:
            raise ValueError("Number of samples in the input and target "
                             "datasets are different")



from __future__ import division

import math

import tensorflow as tf

from neupy.utils import as_tuple, tf_repeat
from neupy.core.properties import TypedListProperty, ChoiceProperty, Property
from neupy.exceptions import LayerConnectionError
from .base import BaseLayer
from .convolutions import Spatial2DProperty


__all__ = ('MaxPooling', 'AveragePooling', 'Upscale', 'GlobalPooling')


def pooling_output_shape(dimension_size, pool_size, padding, stride):
    """
    Computes output shape for pooling operation.

    Parameters
    ----------
    dimension_size : int
        Size of the dimension. Typically it's image's
        weight or height.

    filter_size : int
        Size of the pooling filter.

    padding : int
        Size of the zero-padding.

    stride : int
        Stride size.

    Returns
    -------
    int
    """
    if dimension_size is None:
        return None

    if padding == 'SAME':
        return math.ceil(dimension_size / stride)

    elif padding == 'VALID':
        return math.ceil((dimension_size - pool_size + 1) / stride)

    raise ValueError("`{!r}` is unknown convolution's padding value"
                     "".format(padding))


class BasePooling(BaseLayer):
    """
    Base class for the pooling layers.

    Parameters
    ----------
    size : tuple with 2 integers
        Factor by which to downscale (vertical, horizontal).
        (2, 2) will halve the image in each dimension.

    stride : tuple or int.
        Stride size, which is the number of shifts over
        rows/cols to get the next pool region. If stride is
        None, it is considered equal to ds (no overlap on
        pooling regions).

    padding : {{`VALID`, `SAME`}}
        (pad_h, pad_w), pad zeros to extend beyond four borders of
        the images, pad_h is the size of the top and bottom margins,
        and pad_w is the size of the left and right margins.

    {BaseLayer.Parameters}

    Methods
    -------
    {BaseLayer.Methods}

    Attributes
    ----------
    {BaseLayer.Attributes}
    """
    size = TypedListProperty(required=True, element_type=int)
    stride = Spatial2DProperty(default=None)
    padding = ChoiceProperty(default='VALID', choices=('SAME', 'VALID'))
    pooling_type = None

    def __init__(self, size, **options):
        super(BasePooling, self).__init__(size=size, **options)

    def validate(self, input_shape):
        if len(input_shape) != 3:
            raise LayerConnectionError(
                "Pooling layer expects an input with 3 "
                "dimensions, got {} with shape {}"
                "".format(len(input_shape), input_shape)
            )

    @property
    def output_shape(self):
        if self.input_shape is None:
            return None

        rows, cols, n_kernels = self.input_shape
        row_filter_size, col_filter_size = self.size

        stride = self.size if self.stride is None else self.stride
        row_stride, col_stride = stride

        output_rows = pooling_output_shape(
            rows, row_filter_size, self.padding, row_stride)

        output_cols = pooling_output_shape(
            cols, col_filter_size, self.padding, col_stride)

        # In python 2, we can get float number after rounding procedure
        # and it might break processing in the subsequent layers.
        return (int(output_rows), int(output_cols), n_kernels)

    def output(self, input_value):
        return tf.nn.pool(
            input_value,
            self.size,
            pooling_type=self.pooling_type,
            padding=self.padding,
            strides=self.stride or self.size,
            data_format="NHWC",
        )

    def __repr__(self):
        return '{name}({size})'.format(
            name=self.__class__.__name__,
            size=self.size,
        )


class MaxPooling(BasePooling):
    """
    Maximum pooling layer.

    Parameters
    ----------
    {BasePooling.Parameters}

    Methods
    -------
    {BasePooling.Methods}

    Attributes
    ----------
    {BasePooling.Attributes}

    Examples
    --------
    2D pooling

    >>> from neupy import layers
    >>>
    >>> network = layers.join(
    ...     layers.Input((10, 10, 3)),
    ...     layers.MaxPooling((2, 2)),
    ... )
    >>> network.output_shape
    (3, 5, 5)

    1D pooling

    >>> from neupy import layers
    >>>
    >>> network = layers.join(
    ...     layers.Input((30, 10)),
    ...     layers.Reshape((10, 1, 30)),
    ...     layers.MaxPooling((2, 1)),
    ... )
    >>> network.output_shape
    (10, 15, 1)
    """
    pooling_type = 'MAX'


class AveragePooling(BasePooling):
    """
    Average pooling layer.

    Parameters
    ----------
    {BasePooling.Parameters}

    Methods
    -------
    {BasePooling.Methods}

    Attributes
    ----------
    {BasePooling.Attributes}

    Examples
    --------
    2D pooling

    >>> from neupy import layers
    >>>
    >>> network = layers.join(
    ...     layers.Input((10, 10, 3)),
    ...     layers.AveragePooling((2, 2)),
    ... )
    >>> network.output_shape
    (3, 5, 5)

    1D pooling

    >>> from neupy import layers
    >>>
    >>> network = layers.join(
    ...     layers.Input((30, 10)),
    ...     layers.Reshape((10, 1, 30)),
    ...     layers.AveragePooling((2, 1)),
    ... )
    >>> network.output_shape
    (10, 15, 1)
    """
    pooling_type = 'AVG'


class ScaleFactorProperty(TypedListProperty):
    """
    Defines sclaing factor for the Upscale layer.

    Parameters
    ----------
    {TypedListProperty.Parameters}
    """
    expected_type = (tuple, int)

    def __set__(self, instance, value):
        if isinstance(value, int):
            value = as_tuple(value, value)
        super(ScaleFactorProperty, self).__set__(instance, value)

    def validate(self, value):
        if any(element <= 0 for element in value):
            raise ValueError("Scale factor property accepts only positive "
                             "integer numbers.")
        super(ScaleFactorProperty, self).validate(value)


class Upscale(BaseLayer):
    """
    Upscales input over two axis (height and width).

    Parameters
    ----------
    scale : int or tuple with two int
        Scaling factor for the input value. In the tuple first
        parameter identifies scale of the height and the second
        one of the width.

    {BaseLayer.Parameters}

    Methods
    -------
    {BaseLayer.Methods}

    Attributes
    ----------
    {BaseLayer.Attributes}

    Examples
    --------
    >>> from neupy.layers import *
    >>> network = Input((10, 10, 3)) > Upscale((2, 2))
    >>> network.output_shape
    (3, 20, 20)
    """
    scale = ScaleFactorProperty(required=True, n_elements=2)

    def __init__(self, scale, **options):
        super(Upscale, self).__init__(scale=scale, **options)

    def validate(self, input_shape):
        if len(input_shape) != 3:
            raise LayerConnectionError(
                "Upscale layer should have an input value with "
                "3 feature dimensions (channel, height, width)")

    @property
    def output_shape(self):
        if self.input_shape is None:
            return None

        height, width, channel = self.input_shape
        height_scale, width_scale = self.scale

        return (height_scale * height, width_scale * width, channel)

    def output(self, input_value):
        return tf_repeat(input_value, as_tuple(1, self.scale, 1))


class GlobalPooling(BaseLayer):
    """
    Global pooling layer.

    Parameters
    ----------
    function : callable
        Function that aggregates over dimensions.
        Defaults to ``tensorflow.reduce_mean``.

        .. code-block:: python

            def agg_func(x, axis=None):
                pass

    {BaseLayer.Parameters}

    Methods
    -------
    {BaseLayer.Methods}

    Attributes
    ----------
    {BaseLayer.Attributes}

    Examples
    --------
    >>> from neupy.layers import *
    >>> network = Input((4, 4, 16)) > GlobalPooling()
    >>> network.output_shape
    (16,)
    """
    function = Property(default=tf.reduce_mean)

    @property
    def output_shape(self):
        if self.input_shape is not None:
            return as_tuple(self.input_shape[-1])

    def output(self, input_value):
        ndims = len(input_value.shape)

        if ndims in (1, 2):
            return input_value

        # All dimensions except first and last
        agg_axis = range(1, ndims - 1)
        return self.function(input_value, axis=list(agg_axis))



import numpy as np
import matplotlib.pyplot as plt
from sklearn import datasets
from neupy import algorithms, layers, environment


environment.reproducible()

mnist = datasets.fetch_mldata('MNIST original')

data = mnist.data
data = (data / 255.).astype(np.float32)

np.random.shuffle(data)
x_train, x_test = data[:60000], data[60000:]

autoencoder = algorithms.Momentum(
    [
        layers.Input(784),
        layers.GaussianNoise(mean=0.5, std=0.1),
        layers.Sigmoid(100),
        layers.Sigmoid(784),
    ],
    step=0.25,
    verbose=True,
    momentum=0.99,
    nesterov=True,
    error='rmse',
)
autoencoder.train(x_train, x_train, x_test, x_test, epochs=100)

n_samples = 4
image_vectors = x_test[:n_samples, :]
images = image_vectors * 255.
predicted_images = autoencoder.predict(image_vectors)
predicted_images = predicted_images * 255.

# Compare real and reconstructed images
fig, axes = plt.subplots(4, 2, figsize=(12, 8))
iterator = zip(axes, images, predicted_images)

for (left_ax, right_ax), real_image, predicted_image in iterator:
    real_image = real_image.reshape((28, 28))
    predicted_image = predicted_image.reshape((28, 28))

    left_ax.imshow(real_image, cmap=plt.cm.binary)
    right_ax.imshow(predicted_image, cmap=plt.cm.binary)

plt.show()



import tensorflow as tf

from neupy.core.properties import ProperFractionProperty, NumberProperty
from .base import MinibatchGradientDescent


__all__ = ('RMSProp',)


class RMSProp(MinibatchGradientDescent):
    """
    RMSProp algorithm.

    Parameters
    ----------
    decay : float
        Decay rate. Value need to be between ``0`` and ``1``.
        Defaults to ``0.95``.

    epsilon : float
        Value need to be greater than ``0``. Defaults to ``1e-7``.

    {MinibatchGradientDescent.Parameters}

    Attributes
    ----------
    {MinibatchGradientDescent.Attributes}

    Methods
    -------
    {MinibatchGradientDescent.Methods}

    Examples
    --------
    >>> import numpy as np
    >>> from neupy import algorithms
    >>>
    >>> x_train = np.array([[1, 2], [3, 4]])
    >>> y_train = np.array([[1], [0]])
    >>>
    >>> mnet = algorithms.RMSProp((2, 3, 1))
    >>> mnet.train(x_train, y_train)
    """
    decay = ProperFractionProperty(default=0.95)
    epsilon = NumberProperty(default=1e-7, minval=0)

    def init_train_updates(self):
        updates = []
        step = self.variables.step

        for layer, parameter, gradient in self.iter_params_and_grads():
            prev_mean_squred_grad = tf.Variable(
                tf.zeros(parameter.shape),
                name="{}/prev-mean-squared-grad".format(parameter.op.name),
                dtype=tf.float32,
            )

            mean_squred_grad = (
                self.decay * prev_mean_squred_grad +
                (1. - self.decay) * tf.pow(gradient, 2)
            )
            parameter_delta = gradient / tf.sqrt(
                mean_squred_grad + self.epsilon)

            updates.extend([
                (prev_mean_squred_grad, mean_squred_grad),
                (parameter, parameter - step * parameter_delta),
            ])

        return updates



import math

import numpy as np
import tensorflow as tf

from neupy.utils import asfloat


__all__ = ('fmin_golden_search',)


def interval_location(f, minstep=1e-5, maxstep=50., maxiter=1024):
    """
    Identify interval where potentialy could be optimal step.

    Parameters
    ----------
    f : func

    minstep : float
        Defaults to ``1e-5``.

    maxstep : float
        Defaults to ``50``.

    maxiter : int
        Defaults to ``1024``.

    tol : float
        Defaults to ``1e-5``.

    Returns
    -------
    float
        Right bound of interval where could be optimal step in
        specified direction. In case if there is no such direction
        function return ``maxstep`` instead.
    """
    with tf.name_scope('interval-location'):
        def find_right_bound(_, prev_func_output, step):
            func_output = f(step)
            step = tf.where(
                tf.greater(prev_func_output, func_output),
                tf.minimum(2. * step, maxstep),
                step
            )
            continue_searching = tf.logical_and(
                tf.greater_equal(prev_func_output, func_output),
                step < maxstep,
            )
            return [continue_searching, func_output, step]

        _, _, step = tf.while_loop(
            cond=lambda continue_searching, *args: continue_searching,
            body=find_right_bound,
            loop_vars=[
                True,
                tf.constant(asfloat(np.inf)),
                tf.constant(asfloat(minstep)),
            ],
            back_prop=False,
            maximum_iterations=maxiter,
        )
        return step


def golden_search(f, maxstep=50, maxiter=1024, tol=1e-5):
    """
    Identify best step for function in specific direction.

    Parameters
    ----------
    f : func

    maxstep : float
        Defaults to ``50``.

    maxiter : int
        Defaults to ``1024``.

    tol : float
        Defaults to ``1e-5``.

    Returns
    -------
    float
        Identified optimal step.
    """
    with tf.name_scope('golden-search'):
        golden_ratio = asfloat((math.sqrt(5) - 1) / 2)
        tol = asfloat(tol)

        def interval_reduction(a, b, c, d):
            fc = f(c)
            fd = f(d)

            is_c_smaller_than_d = tf.less(fc, fd)

            new_a = tf.where(is_c_smaller_than_d, a, c)
            new_b = tf.where(is_c_smaller_than_d, d, b)
            new_c = tf.where(
                is_c_smaller_than_d, d - golden_ratio * (d - a), d)
            new_d = tf.where(
                is_c_smaller_than_d, c, c + golden_ratio * (b - c))

            return [new_a, new_b, new_c, new_d]

        a = tf.constant(asfloat(0))
        b = maxstep
        c = b - golden_ratio * (b - a)
        d = a + golden_ratio * (b - a)

        a, b, c, d = tf.while_loop(
            cond=lambda a, b, c, d: tf.greater(tf.abs(c - d), tol),
            body=interval_reduction,
            loop_vars=[a, b, c, d],
            back_prop=False,
            maximum_iterations=maxiter,
        )
        return (a + b) / 2


def fmin_golden_search(f, minstep=1e-5, maxstep=50., maxiter=1024, tol=1e-5):
    """
    Minimize scalar function using Golden Search.

    Parameters
    ----------
    f : func
        Function that needs to be minimized. Function has
        to return the scalar.

        .. code-block:: python

            def f(x):
                return x ** 2

    minstep : float
        Minimum step value. Defaults to ``1e-5``.

    maxstep : float
        Maximum step size. Defaults to ``50``.

    maxiter : int
        Maximum number of itrations. Defaults to ``1024``.

    tol : float
        Defaults to ``1e-5``.

    Returns
    -------
    object
        Returns the Tensorfow instance that finally should produce
        best possbile step for specified function.
    """
    params = (
        ('maxiter', maxiter),
        ('minstep', minstep),
        ('maxstep', maxstep),
        ('tol', tol),
    )

    for param_name, param_value in params:
        if param_value <= 0:
            raise ValueError("Parameter `{}` should be greater than zero."
                             "".format(param_name))

    if minstep >= maxstep:
        raise ValueError("`minstep` should be smaller than `maxstep`")

    maxstep = interval_location(f, minstep, maxstep, maxiter)
    best_step = golden_search(f, maxstep, maxiter, tol)

    return best_step



import numpy as np
from sklearn import datasets
import matplotlib.pyplot as plt
from neupy import algorithms, layers, environment


environment.reproducible()

mnist = datasets.fetch_mldata('MNIST original')
data = (mnist.data / 255.).astype(np.float32)

np.random.shuffle(data)
x_train, x_test = data[:60000], data[60000:]
x_train_4d = x_train.reshape((60000, 28, 28, 1))
x_test_4d = x_test.reshape((10000, 28, 28, 1))

conv_autoencoder = algorithms.Momentum(
    [
        layers.Input((28, 28, 1)),

        layers.Convolution((3, 3, 16)) > layers.Relu(),
        layers.Convolution((3, 3, 16)) > layers.Relu(),
        layers.MaxPooling((2, 2)),

        layers.Convolution((3, 3, 32)) > layers.Relu(),
        layers.MaxPooling((2, 2)),

        layers.Reshape(),

        layers.Relu(128),
        layers.Relu(16),

        # Notice that in the decoder every operation reverts back changes
        # from the encoder layer. Upscale replaces MaxPooling and
        # Convolutional layer without padding replaced with large padding
        # that increase size of the image.
        layers.Relu(128),

        # 800 is a shape that we got after we reshaped our image in the
        # Reshape layer
        layers.Relu(800),

        layers.Reshape((5, 5, 32)),

        # Upscaling layer reverts changes from the max pooling layer
        layers.Upscale((2, 2)),

        # If convolution operation in first layers with zero padding reduces
        # size of the image, then convolution with padding=2 increases size
        # of the image. It just does the opposite to the previous convolution
        layers.Convolution((3, 3, 16), padding=2) > layers.Relu(),

        layers.Upscale((2, 2)),
        layers.Convolution((3, 3, 16), padding=2) > layers.Relu(),
        layers.Convolution((3, 3, 1), padding=2) > layers.Sigmoid(),

        # We have to convert 4d tensor to the 2d in order to be
        # able to compute RMSE.
        layers.Reshape(),
    ],

    step=0.02,
    momentum=0.9,
    batch_size=128,
    error='rmse',

    shuffle_data=True,
    verbose=True,

    decay_rate=0.01,
    addons=[algorithms.WeightDecay],
)
conv_autoencoder.architecture()
conv_autoencoder.train(x_train_4d, x_train, x_test_4d, x_test, epochs=15)

n_samples = 6
images = x_test[:n_samples] * 255.
predicted_images = conv_autoencoder.predict(x_test_4d[:n_samples])
predicted_images = predicted_images * 255.

# Compare real and reconstructed images
fig, axes = plt.subplots(n_samples, 2, figsize=(12, 8))
iterator = zip(axes, images, predicted_images)

for (left_ax, right_ax), real_image, predicted_image in iterator:
    real_image = real_image.reshape((28, 28))
    predicted_image = predicted_image.reshape((28, 28))

    left_ax.imshow(real_image, cmap=plt.cm.binary)
    right_ax.imshow(predicted_image, cmap=plt.cm.binary)

plt.show()



from abc import abstractmethod

from neupy.utils import preformat_value
from neupy.core.logs import Verbose
from .config import ConfigurableABC


__all__ = ('BaseSkeleton',)


class BaseSkeleton(ConfigurableABC, Verbose):
    """
    Base class for neural network algorithms.

    Methods
    -------
    fit(\*args, \*\*kwargs)
        Alias to the ``train`` method.

    predict(input_data)
        Predicts output for the specified input.
    """
    @abstractmethod
    def train(self, input_data, target_data):
        raise NotImplementedError

    def predict(self, input_data):
        raise NotImplementedError

    def transform(self, input_data):
        return self.predict(input_data)

    def fit(self, X, y, *args, **kwargs):
        self.train(X, y, *args, **kwargs)
        return self

    def class_name(self):
        return self.__class__.__name__

    def repr_options(self):
        options = []
        for option_name in self.options:
            option_value = getattr(self, option_name)
            option_value = preformat_value(option_value)

            option_repr = "{}={}".format(option_name, option_value)
            options.append(option_repr)

        return ', '.join(options)

    def __repr__(self):
        class_name = self.class_name()
        available_options = self.repr_options()
        return "{}({})".format(class_name, available_options)



import tensorflow as tf

from neupy.core.properties import ProperFractionProperty, NumberProperty
from .base import MinibatchGradientDescent


__all__ = ('Adadelta',)


class Adadelta(MinibatchGradientDescent):
    """
    Adadelta algorithm.

    Parameters
    ----------
    decay : float
        Decay rate. Value need to be between ``0``
        and ``1``. Defaults to ``0.95``.

    epsilon : float
        Value need to be greater than ``0``. Defaults to ``1e-7``.

    step : float
        Learning rate, defaults to ``1.0``. Original paper doesn't have
        learning rate specified in the paper. Step value equal to ``1.0``
        allow to achive the same effect, since multiplication by one won't
        have any effect on the update.

    {MinibatchGradientDescent.batch_size}

    {GradientDescent.addons}

    {ConstructibleNetwork.connection}

    {ConstructibleNetwork.error}

    {BaseNetwork.show_epoch}

    {BaseNetwork.shuffle_data}

    {BaseNetwork.epoch_end_signal}

    {BaseNetwork.train_end_signal}

    Attributes
    ----------
    {MinibatchGradientDescent.Attributes}

    Methods
    -------
    {MinibatchGradientDescent.Methods}

    Examples
    --------
    >>> import numpy as np
    >>> from neupy import algorithms
    >>>
    >>> x_train = np.array([[1, 2], [3, 4]])
    >>> y_train = np.array([[1], [0]])
    >>>
    >>> mnet = algorithms.Adadelta((2, 3, 1))
    >>> mnet.train(x_train, y_train)

    References
    ----------
    [1] Matthew D. Zeiler,
        ADADELTA: An Adaptive Learning Rate Method
        https://arxiv.org/pdf/1212.5701.pdf
    """
    step = NumberProperty(default=1.0, minval=0)
    decay = ProperFractionProperty(default=0.95)
    epsilon = NumberProperty(default=1e-7, minval=0)

    def init_train_updates(self):
        updates = []
        step = self.variables.step
        epsilon = self.epsilon

        for layer, parameter, gradient in self.iter_params_and_grads():
            prev_mean_squred_grad = tf.Variable(
                tf.zeros(parameter.shape),
                name="{}/prev-mean-squred-grad".format(parameter.op.name),
                dtype=tf.float32,
            )
            prev_mean_squared_update = tf.Variable(
                tf.zeros(parameter.shape),
                name="{}/prev-mean-squred-update".format(parameter.op.name),
                dtype=tf.float32,
            )

            mean_squred_grad = (
                self.decay * prev_mean_squred_grad +
                (1 - self.decay) * gradient ** 2
            )
            parameter_delta = gradient * (
                tf.sqrt(prev_mean_squared_update + epsilon) /
                tf.sqrt(mean_squred_grad + epsilon)
            )
            mean_squared_update = (
                self.decay * prev_mean_squared_update +
                (1 - self.decay) * parameter_delta ** 2
            )

            updates.extend([
                (prev_mean_squred_grad, mean_squred_grad),
                (prev_mean_squared_update, mean_squared_update),
                (parameter, parameter - step * parameter_delta),
            ])

        return updates



from neupy.utils import asfloat
from neupy.layers.utils import iter_parameters
from neupy.core.properties import BoundedProperty
from .base import WeightUpdateConfigurable


__all__ = ('WeightDecay',)


class WeightDecay(WeightUpdateConfigurable):
    """
    Weight decay algorithm penalizes large weights. Also known as
    L2-regularization.

    Parameters
    ----------
    decay_rate : float
        Controls training penalties during the parameter updates.
        The larger the value the stronger effect regularization
        has during the training. Defaults to ``0.1``.

    Warns
    -----
    {WeightUpdateConfigurable.Warns}

    Examples
    --------
    >>> from neupy import algorithms
    >>> bpnet = algorithms.GradientDescent(
    ...     (2, 4, 1),
    ...     step=0.1,
    ...     decay_rate=0.1,
    ...     addons=[algorithms.WeightDecay]
    ... )

    See Also
    --------
    :network:`WeightElimination`
    """
    decay_rate = BoundedProperty(default=0.1, minval=0)

    def init_train_updates(self):
        original_updates = super(WeightDecay, self).init_train_updates()
        parameters = [param for _, _, param in iter_parameters(self.layers)]
        modified_updates = []

        step = self.variables.step
        decay_rate = asfloat(self.decay_rate)

        for parameter, updated in original_updates:
            if parameter in parameters:
                updated -= step * decay_rate * parameter
            modified_updates.append((parameter, updated))

        return modified_updates



import numpy as np
import tensorflow as tf

from neupy import init
from neupy.utils import AttributeKeyDict, as_tuple, tensorflow_session
from neupy.exceptions import LayerConnectionError
from neupy.core.properties import (IntProperty, Property, NumberProperty,
                                   ParameterProperty)
from .base import BaseLayer


__all__ = ('LSTM', 'GRU')


def clip_gradient(value, clip_value):
    if not hasattr(clip_gradient, 'added_gradients'):
        clip_gradient.added_gradients = set()

    session = tensorflow_session()
    graph = session.graph
    operation_name = "ClipGradient-" + str(clip_value)

    if operation_name not in clip_gradient.added_gradients:
        # Make sure that we won't create the same operation twise.
        # Otherwise tensorflow will trigger an exception.
        @tf.RegisterGradient(operation_name)
        def clip_gradient_grad(op, grad):
            return tf.clip_by_value(grad, -clip_value, clip_value)

        clip_gradient.added_gradients.add(operation_name)

    with graph.gradient_override_map({"Identity": operation_name}):
        return tf.identity(value)


def unroll_scan(fn, sequence, outputs_info):
    """
    Helper function to unroll for loops. Can be used to unroll
    ``tensorflow.scan``.

    Parameters
    ----------
    fn : function
        Function that defines calculations at each step.

    sequences : TensorVariable or list of TensorVariables
        List of TensorVariable with sequence data. The function iterates
        over the first dimension of each TensorVariable.

    outputs_info : list of TensorVariables
        List of tensors specifying the initial values for each recurrent
        value.

    Returns
    -------
    List of TensorVariables. Each element in the list gives
    the recurrent values at each time step.
    """
    with tf.name_scope('unroll-scan'):
        outputs = []
        prev_vals = outputs_info

        for entity in tf.unstack(sequence):
            output = fn(prev_vals, entity)
            outputs.append(output[-1])
            prev_vals = output

        return tf.stack(outputs)


class MultiCallableProperty(ParameterProperty):
    expected_type = as_tuple(dict)

    def __set__(self, instance, value):
        self.validate(value)

        default_value = self.default.copy()
        default_value.update(value)
        value = default_value

        value = AttributeKeyDict(value)
        instance.__dict__[self.name] = value

    def validate(self, value):
        if not isinstance(value, self.expected_type):
            raise TypeError(
                "Parameter `{}` should be a dictionary, "
                "got `{!r}`".format(self.name, type(value)))

        for key, func in value.items():
            if not callable(func):
                raise ValueError(
                    "Values for the `{}` parameter should be "
                    "callable objects, got value `{!r}` for the "
                    "`{}` key".format(self.name, func, key))


class BaseRNNLayer(BaseLayer):
    """
    Base class for the recurrent layers

    Parameters
    ----------
    size : int
        Number of hidden units in the layer.

    only_return_final : bool
        If ``True``, only return the final sequential output
        (e.g. for tasks where a single target value for the entire
        sequence is desired). In this case, Tensorfow makes an
        optimization which saves memory. Defaults to ``True``.

    {BaseLayer.Parameters}
    """
    size = IntProperty(minval=1)
    only_return_final = Property(default=True, expected_type=bool)

    def __init__(self, size, **kwargs):
        super(BaseRNNLayer, self).__init__(size=size, **kwargs)

    def validate(self, input_shape):
        n_input_dims = len(input_shape) + 1  # +1 for batch dimension
        clsname = self.__class__.__name__

        if n_input_dims != 3:
            raise LayerConnectionError(
                "{} layer was expected input with three dimensions, "
                "but got input with {} dimensions instead"
                "".format(clsname, n_input_dims))

    @property
    def output_shape(self):
        if self.only_return_final:
            return as_tuple(self.size)

        n_time_steps = self.input_shape[0]
        return as_tuple(n_time_steps, self.size)


class LSTM(BaseRNNLayer):
    """
    Long Short Term Memory (LSTM) Layer.

    Parameters
    ----------
    {BaseRNNLayer.size}

    input_weights : Initializer, ndarray
        Weight parameters for input connection.
        Defaults to :class:`XavierUniform() <neupy.init.XavierUniform>`.

    hidden_weights : Initializer, ndarray
        Weight parameters for hidden connection.
        Defaults to :class:`XavierUniform() <neupy.init.XavierUniform>`.

    cell_weights : Initializer, ndarray
        Weight parameters for cell connection. Require only when
        ``peepholes=True`` otherwise it will be ignored.
        Defaults to :class:`XavierUniform() <neupy.init.XavierUniform>`.

    bias : Initializer, ndarray
        Bias parameters for all gates.
        Defaults to :class:`Constant(0) <neupy.init.Constant>`.

    activation_functions : dict, callable
        Activation functions for different gates. Defaults to:

        .. code-block:: python

            # import tensorflow as tf
            dict(
                ingate=tf.nn.sigmoid,
                forgetgate=tf.nn.sigmoid,
                outgate=tf.nn.sigmoid,
                cell=tf.tanh,
            )

        If application requires modification to only one parameter
        then it's better to specify the one that you need to modify
        and ignore other parameters

        .. code-block:: python

            dict(ingate=tf.tanh)

        Other parameters like ``forgetgate`` or ``outgate`` will be
        equal to their default values.

    learn_init : bool
        If ``True``, make ``cell_init`` and ``hidden_init`` trainable
        variables. Defaults to ``False``.

    cell_init : array-like, Tensorfow variable, scalar or Initializer
        Initializer for initial cell state (:math:`c_0`).
        Defaults to :class:`Constant(0) <neupy.init.Constant>`.

    hidden_init : array-like, Tensorfow variable, scalar or Initializer
        Initializer for initial hidden state (:math:`h_0`).
        Defaults to :class:`Constant(0) <neupy.init.Constant>`.

    backwards : bool
        If ``True``, process the sequence backwards and then reverse the
        output again such that the output from the layer is always
        from :math:`x_1` to :math:`x_n`. Defaults to ``False``

    {BaseRNNLayer.only_return_final}

    peepholes : bool
        If ``True``, the LSTM uses peephole connections.
        When ``False``, cell parameters  are ignored.
        Defaults to ``False``.

    unroll_scan : bool
        If ``True`` the recursion is unrolled instead of using scan.
        For some graphs this gives a significant speed up but it
        might also consume more memory. When ``unroll_scan=True``,
        backpropagation always includes the full sequence, so
        ``n_gradient_steps`` must be set to ``-1`` and the input
        sequence length must be known at compile time (i.e.,
        cannot be given as ``None``). Defaults to ``False``.

    gradient_clipping : float or int
        If nonzero, the gradient messages are clipped to the
        given value during the backward pass. Defaults to ``0``.

    {BaseLayer.Parameters}

    Notes
    -----
    Code was adapted from the
    `Lasagne <https://github.com/Lasagne/Lasagne>`_ library.

    Examples
    --------

    Sequence classification

    .. code-block:: python

        from neupy import layers, algorithms

        n_time_steps = 40
        n_categories = 20
        embedded_size = 10

        network = algorithms.RMSProp(
            [
                layers.Input(n_time_steps),
                layers.Embedding(n_categories, embedded_size),
                layers.LSTM(20),
                layers.Sigmoid(1),
            ]
        )
    """
    input_weights = ParameterProperty(default=init.XavierUniform())
    hidden_weights = ParameterProperty(default=init.XavierUniform())
    cell_weights = ParameterProperty(default=init.XavierUniform())
    biases = ParameterProperty(default=init.Constant(0))

    activation_functions = MultiCallableProperty(
        default=dict(
            ingate=tf.nn.sigmoid,
            forgetgate=tf.nn.sigmoid,
            outgate=tf.nn.sigmoid,
            cell=tf.tanh,
        )
    )

    learn_init = Property(default=False, expected_type=bool)
    cell_init = ParameterProperty(default=init.Constant(0))
    hidden_init = ParameterProperty(default=init.Constant(0))

    unroll_scan = Property(default=False, expected_type=bool)
    backwards = Property(default=False, expected_type=bool)
    peepholes = Property(default=False, expected_type=bool)
    gradient_clipping = NumberProperty(default=0, minval=0)

    def initialize(self):
        super(LSTM, self).initialize()
        n_inputs = np.prod(self.input_shape[1:])

        # If peephole (cell to gate) connections were enabled, initialize
        # peephole connections.  These are elementwise products with the cell
        # state, so they are represented as vectors.
        if self.peepholes:
            self.weight_cell_to_ingate = self.add_parameter(
                value=self.cell_weights,
                name='weight_cell_to_ingate',
                shape=(self.size,))
            self.weight_cell_to_forgetgate = self.add_parameter(
                value=self.cell_weights,
                name='weight_cell_to_forgetgate',
                shape=(self.size,))
            self.weight_cell_to_outgate = self.add_parameter(
                value=self.cell_weights,
                name='weight_cell_to_outgate',
                shape=(self.size,))

        self.input_weights = self.add_parameter(
            value=self.input_weights,
            name='input_weights',
            shape=(n_inputs, 4 * self.size),
        )
        self.hidden_weights = self.add_parameter(
            value=self.hidden_weights,
            name='hidden_weights',
            shape=(self.size, 4 * self.size),
        )
        self.biases = self.add_parameter(
            value=self.biases, name='biases',
            shape=(4 * self.size,),
        )

        # Initialization parameters
        self.add_parameter(
            value=self.cell_init,
            shape=(1, self.size),
            name="cell_init",
            trainable=self.learn_init,
        )
        self.add_parameter(
            value=self.hidden_init,
            shape=(1, self.size),
            name="hidden_init",
            trainable=self.learn_init,
        )

    def output(self, input_value):
        # Because scan iterates over the first dimension we
        # dimshuffle to (n_time_steps, n_batch, n_features)
        input_value = tf.transpose(input_value, [1, 0, 2])
        input_shape = tf.shape(input_value)
        n_batch = input_shape[1]

        def one_lstm_step(states, input_n):
            with tf.name_scope('lstm-cell'):
                cell_previous, hid_previous = states
                input_n = tf.matmul(input_n, self.input_weights) + self.biases

                # Calculate gates pre-activations and slice
                gates = input_n + tf.matmul(hid_previous, self.hidden_weights)

                # Clip gradients
                if self.gradient_clipping != 0:
                    gates = clip_gradient(gates, self.gradient_clipping)

                # Extract the pre-activation gate values
                ingate, forgetgate, cell_input, outgate = tf.split(
                    gates, 4, axis=1)

                if self.peepholes:
                    # Compute peephole connections
                    ingate += cell_previous * self.weight_cell_to_ingate
                    forgetgate += (
                        cell_previous * self.weight_cell_to_forgetgate)

                # Apply nonlinearities
                ingate = self.activation_functions.ingate(ingate)
                forgetgate = self.activation_functions.forgetgate(forgetgate)
                cell_input = self.activation_functions.cell(cell_input)

                # Compute new cell value
                cell = forgetgate * cell_previous + ingate * cell_input

                if self.peepholes:
                    outgate += cell * self.weight_cell_to_outgate

                outgate = self.activation_functions.outgate(outgate)

                # Compute new hidden unit activation
                hid = outgate * tf.tanh(cell)
                return [cell, hid]

        cell_init = tf.tile(self.cell_init, (n_batch, 1))
        hidden_init = tf.tile(self.hidden_init, (n_batch, 1))
        sequence = input_value

        if self.backwards:
            sequence = tf.reverse(sequence, axis=[0])

        if self.unroll_scan:
            # Explicitly unroll the recurrence instead of using scan
            hid_out = unroll_scan(
                fn=one_lstm_step,
                sequence=sequence,
                outputs_info=[cell_init, hidden_init],
            )
        else:
            _, hid_out = tf.scan(
                fn=one_lstm_step,
                elems=input_value,
                initializer=[cell_init, hidden_init],
                name='lstm-scan',
            )

        # When it is requested that we only return the final sequence step,
        # we need to slice it out immediately after scan is applied
        if self.only_return_final:
            return hid_out[-1]

        # if scan is backward reverse the output
        if self.backwards:
            hid_out = tf.reverse(hid_out, axis=[0])

        # dimshuffle back to (n_batch, n_time_steps, n_features))
        hid_out = tf.transpose(hid_out, [1, 0, 2])

        return hid_out


class GRU(BaseRNNLayer):
    """
    Gated Recurrent Unit (GRU) Layer.

    Parameters
    ----------
    {BaseRNNLayer.size}

    input_weights : Initializer, ndarray
        Weight parameters for input connection.
        Defaults to :class:`XavierUniform() <neupy.init.XavierUniform>`.

    hidden_weights : Initializer, ndarray
        Weight parameters for hidden connection.
        Defaults to :class:`XavierUniform() <neupy.init.XavierUniform>`.

    bias : Initializer, ndarray
        Bias parameters for all gates.
        Defaults to :class:`Constant(0) <neupy.init.Constant>`.

    activation_functions : dict, callable
        Activation functions for different gates. Defaults to:

        .. code-block:: python

            # import tensorflow as tf
            dict(
                resetgate=tf.nn.sigmoid,
                updategate=tf.nn.sigmoid,
                hidden_update=tf.tanh,
            )

        If application requires modification to only one parameter
        then it's better to specify the one that you need to modify
        and ignore other parameters

        .. code-block:: python

            dict(resetgate=tf.tanh)

        Other parameters like ``updategate`` or ``hidden_update``
        will be equal to their default values.

    learn_init : bool
        If ``True``, make ``hidden_init`` trainable variable.
        Defaults to ``False``.

    hidden_init : array-like, Tensorfow variable, scalar or Initializer
        Initializer for initial hidden state (:math:`h_0`).
        Defaults to :class:`Constant(0) <neupy.init.Constant>`.

    {BaseRNNLayer.only_return_final}

    backwards : bool
        If ``True``, process the sequence backwards and then reverse the
        output again such that the output from the layer is always
        from :math:`x_1` to :math:`x_n`. Defaults to ``False``.

    unroll_scan : bool
        If ``True`` the recursion is unrolled instead of using scan.
        For some graphs this gives a significant speed up but it
        might also consume more memory. When ``unroll_scan=True``,
        backpropagation always includes the full sequence, so
        ``n_gradient_steps`` must be set to ``-1`` and the input
        sequence length must be known at compile time (i.e.,
        cannot be given as ``None``). Defaults to ``False``.

    {BaseLayer.Parameters}

    Notes
    -----
    Code was adapted from the
    `Lasagne <https://github.com/Lasagne/Lasagne>`_ library.

    Examples
    --------

    Sequence classification

    .. code-block:: python

        from neupy import layers, algorithms

        n_time_steps = 40
        n_categories = 20
        embedded_size = 10

        network = algorithms.RMSProp(
            [
                layers.Input(n_time_steps),
                layers.Embedding(n_categories, embedded_size),
                layers.GRU(20),
                layers.Sigmoid(1),
            ]
        )
    """
    input_weights = ParameterProperty(default=init.XavierUniform())
    hidden_weights = ParameterProperty(default=init.XavierUniform())
    biases = ParameterProperty(default=init.Constant(0))

    activation_functions = MultiCallableProperty(
        default=dict(
            resetgate=tf.nn.sigmoid,
            updategate=tf.nn.sigmoid,
            hidden_update=tf.tanh,
        )
    )

    learn_init = Property(default=False, expected_type=bool)
    hidden_init = ParameterProperty(default=init.Constant(0))

    backwards = Property(default=False, expected_type=bool)
    unroll_scan = Property(default=False, expected_type=bool)
    gradient_clipping = NumberProperty(default=0, minval=0)

    def initialize(self):
        super(GRU, self).initialize()
        n_inputs = np.prod(self.input_shape[1:])

        self.input_weights = self.add_parameter(
            value=self.input_weights,
            name='input_weights',
            shape=(n_inputs, 3 * self.size),
        )
        self.hidden_weights = self.add_parameter(
            value=self.hidden_weights,
            name='hidden_weights',
            shape=(self.size, 3 * self.size),
        )
        self.biases = self.add_parameter(
            value=self.biases, name='biases',
            shape=(3 * self.size,),
        )

        self.add_parameter(
            value=self.hidden_init,
            shape=(1, self.size),
            name="hidden_init",
            trainable=self.learn_init
        )

    def output(self, input_value):
        # Because scan iterates over the first dimension we
        # dimshuffle to (n_time_steps, n_batch, n_features)
        input_value = tf.transpose(input_value, [1, 0, 2])
        input_shape = tf.shape(input_value)
        n_batch = input_shape[1]

        # Create single recurrent computation step function
        # input_n is the n'th vector of the input
        def one_gru_step(states, input_n):
            with tf.name_scope('gru-cell'):
                hid_previous, = states
                input_n = tf.matmul(input_n, self.input_weights) + self.biases

                # Compute W_{hr} h_{t - 1}, W_{hu} h_{t - 1},
                # and W_{hc} h_{t - 1}
                hid_input = tf.matmul(hid_previous, self.hidden_weights)

                if self.gradient_clipping != 0:
                    input_n = clip_gradient(input_n, self.gradient_clipping)
                    hid_input = clip_gradient(
                        hid_input, self.gradient_clipping)

                hid_resetgate, hid_updategate, hid_hidden = tf.split(
                    hid_input, 3, axis=1)

                in_resetgate, in_updategate, in_hidden = tf.split(
                    input_n, 3, axis=1)

                # Reset and update gates
                resetgate = self.activation_functions.resetgate(
                    hid_resetgate + in_resetgate)

                updategate = self.activation_functions.updategate(
                    hid_updategate + in_updategate)

                # Compute W_{xc}x_t + r_t \odot (W_{hc} h_{t - 1})
                hidden_update = in_hidden + resetgate * hid_hidden

                if self.gradient_clipping != 0:
                    hidden_update = clip_gradient(
                        hidden_update, self.gradient_clipping)

                hidden_update = self.activation_functions.hidden_update(
                    hidden_update)

                # Compute (1 - u_t)h_{t - 1} + u_t c_t
                return [
                    hid_previous - updategate * (hid_previous - hidden_update)
                ]

        hidden_init = tf.tile(self.hidden_init, (n_batch, 1))
        sequence = input_value

        if self.backwards:
            sequence = tf.reverse(sequence, axis=[0])

        if self.unroll_scan:
            # Explicitly unroll the recurrence instead of using scan
            hid_out = unroll_scan(
                fn=one_gru_step,
                sequence=sequence,
                outputs_info=[hidden_init]
            )
        else:
            hid_out, = tf.scan(
                fn=one_gru_step,
                elems=input_value,
                initializer=[hidden_init],
                name='gru-scan',
            )

        # When it is requested that we only return the final sequence step,
        # we need to slice it out immediately after scan is applied
        if self.only_return_final:
            return hid_out[-1]

        # if scan is backward reverse the output
        if self.backwards:
            hid_out = tf.reverse(hid_out, axis=[0])

        # dimshuffle back to (n_batch, n_time_steps, n_features))
        hid_out = tf.transpose(hid_out, [1, 0, 2])
        return hid_out



import tensorflow as tf

from neupy.utils import dot, function_name_scope
from neupy.core.properties import (ChoiceProperty, NumberProperty,
                                   WithdrawProperty)
from neupy.layers.utils import count_parameters
from neupy.algorithms.utils import (parameter_values, setup_parameter_updates,
                                    make_single_vector)
from .base import GradientDescent
from .quasi_newton import safe_division, WolfeLineSearchForStep


__all__ = ('ConjugateGradient',)


@function_name_scope
def fletcher_reeves(old_g, new_g, delta_w, epsilon=1e-7):
    return safe_division(
        dot(new_g, new_g),
        dot(old_g, old_g),
        epsilon,
    )


@function_name_scope
def polak_ribiere(old_g, new_g, delta_w, epsilon=1e-7):
    return safe_division(
        dot(new_g, new_g - old_g),
        dot(old_g, old_g),
        epsilon,
    )


@function_name_scope
def hentenes_stiefel(old_g, new_g, delta_w, epsilon=1e-7):
    gradient_delta = new_g - old_g
    return safe_division(
        dot(gradient_delta, new_g),
        dot(delta_w, gradient_delta),
        epsilon,
    )


@function_name_scope
def liu_storey(old_g, new_g, delta_w, epsilon=1e-7):
    return -safe_division(
        dot(new_g, new_g - old_g),
        dot(delta_w, old_g),
        epsilon,
    )


@function_name_scope
def dai_yuan(old_g, new_g, delta_w, epsilon=1e-7):
    return safe_division(
        dot(new_g, new_g),
        dot(new_g - old_g, delta_w),
        epsilon,
    )


class ConjugateGradient(WolfeLineSearchForStep, GradientDescent):

    """
    Conjugate Gradient algorithm.

    Parameters
    ----------
    update_function : ``fletcher_reeves``, ``polak_ribiere``,\
    ``hentenes_stiefel``, ``dai_yuan``, ``liu_storey``
        Update function. Defaults to ``fletcher_reeves``.

    epsilon : float
        Ensures computational stability during the devision in
        ``update_function`` when denumerator is very small number.
        Defaults to ``1e-7``.

    {WolfeLineSearchForStep.Parameters}

    {GradientDescent.connection}

    {GradientDescent.error}

    {GradientDescent.show_epoch}

    {GradientDescent.shuffle_data}

    {GradientDescent.epoch_end_signal}

    {GradientDescent.train_end_signal}

    {GradientDescent.verbose}

    {GradientDescent.addons}

    Attributes
    ----------
    {GradientDescent.Attributes}

    Methods
    -------
    {GradientDescent.Methods}

    Examples
    --------
    >>> from sklearn import datasets, preprocessing
    >>> from sklearn.model_selection import train_test_split
    >>> from neupy import algorithms, layers
    >>>
    >>> dataset = datasets.load_boston()
    >>> data, target = dataset.data, dataset.target
    >>>
    >>> data_scaler = preprocessing.MinMaxScaler()
    >>> target_scaler = preprocessing.MinMaxScaler()
    >>>
    >>> x_train, x_test, y_train, y_test = train_test_split(
    ...     data_scaler.fit_transform(data),
    ...     target_scaler.fit_transform(target),
    ...     test_size=0.15
    ... )
    >>>
    >>> cgnet = algorithms.ConjugateGradient(
    ...     connection=[
    ...         layers.Input(13),
    ...         layers.Sigmoid(50),
    ...         layers.Sigmoid(1),
    ...     ],
    ...     update_function='fletcher_reeves',
    ...     verbose=False
    ... )
    >>>
    >>> cgnet.train(x_train, y_train, epochs=100)
    >>> y_predict = cgnet.predict(x_test).round(1)
    >>>
    >>> real = target_scaler.inverse_transform(y_test)
    >>> predicted = target_scaler.inverse_transform(y_predict)

    References
    ----------
    [1] Jorge Nocedal, Stephen J. Wright, Numerical Optimization.
        Chapter 5, Conjugate Gradient Methods, p. 101-133
    """
    epsilon = NumberProperty(default=1e-7, minval=0)
    update_function = ChoiceProperty(
        default='fletcher_reeves',
        choices={
            'fletcher_reeves': fletcher_reeves,
            'polak_ribiere': polak_ribiere,
            'hentenes_stiefel': hentenes_stiefel,
            'liu_storey': liu_storey,
            'dai_yuan': dai_yuan,
        }
    )
    step = WithdrawProperty()

    def init_variables(self):
        super(ConjugateGradient, self).init_variables()
        n_parameters = count_parameters(self.connection)

        self.variables.update(
            prev_delta=tf.Variable(
                tf.zeros([n_parameters]),
                name="conj-grad/prev-delta",
                dtype=tf.float32,
            ),
            prev_gradient=tf.Variable(
                tf.zeros([n_parameters]),
                name="conj-grad/prev-gradient",
                dtype=tf.float32,
            ),
        )

    def init_train_updates(self):
        step = self.variables.step
        epoch = self.variables.epoch
        previous_delta = self.variables.prev_delta
        previous_gradient = self.variables.prev_gradient

        n_parameters = count_parameters(self.connection)
        parameters = parameter_values(self.connection)
        param_vector = make_single_vector(parameters)

        gradients = tf.gradients(self.variables.error_func, parameters)
        full_gradient = make_single_vector(gradients)

        beta = self.update_function(
            previous_gradient, full_gradient, previous_delta, self.epsilon)

        parameter_delta = tf.where(
            tf.equal(tf.mod(epoch, n_parameters), 1),
            -full_gradient,
            -full_gradient + beta * previous_delta
        )

        step = self.find_optimal_step(param_vector, parameter_delta)
        updated_parameters = param_vector + step * parameter_delta
        updates = setup_parameter_updates(parameters, updated_parameters)

        # We have to compute these values first, otherwise
        # parallelization in tensorflow can mix update order
        # and, for example, previous gradient can be equal to
        # current gradient value. It happens because tensorflow
        # try to execute operations in parallel.
        with tf.control_dependencies([full_gradient, parameter_delta]):
            updates.extend([
                previous_gradient.assign(full_gradient),
                previous_delta.assign(parameter_delta),
            ])

        return updates



LEFT = 'left'
RIGHT = 'right'
AND = 'bool'


class GlobalConnectionState(dict):
    def key_to_id(self, key):
        return "[{}] {!r}".format(id(key), key)

    def __setitem__(self, key, value):
        key_id = self.key_to_id(key)
        return super(GlobalConnectionState, self).__setitem__(key_id, value)

    def __getitem__(self, key):
        key_id = self.key_to_id(key)
        return super(GlobalConnectionState, self).__getitem__(key_id)

    def __contains__(self, key):
        key_id = self.key_to_id(key)
        return super(GlobalConnectionState, self).__contains__(key_id)


class EventTracker(GlobalConnectionState):
    def __getitem__(self, key):
        if key not in self:
            self[key] = []
        return super(EventTracker, self).__getitem__(key)

    def add(self, key, event):
        self[key].append(event)

    def activate_on(self, key, events):
        n_events = len(events)
        last_n_events = self[key][-n_events:]
        return last_n_events == events

    def is_gt_before_gt(self, key):
        return self.activate_on(key, [RIGHT, AND, LEFT])

    def is_lt_before_lt(self, key):
        return self.activate_on(key, [LEFT, AND, RIGHT])


class InlineConnection(object):
    left_states = GlobalConnectionState()
    right_states = GlobalConnectionState()
    events = EventTracker()

    def compare(self, left, right):
        self.events.add(left, LEFT)
        self.events.add(right, RIGHT)

        main_left, main_right = left, right

        if self.events.is_gt_before_gt(left):
            left = self.right_states[left]

        if self.events.is_lt_before_lt(right):
            right = self.left_states[right]

        connection = self.connect(left, right)

        self.left_states[main_left] = connection
        self.right_states[main_right] = connection

        return connection

    def __gt__(self, other):
        return self.compare(self, other)

    def __lt__(self, other):
        return self.compare(other, self)

    def __nonzero__(self):
        # For python 2 compatibility
        return self.__bool__()

    def __bool__(self):
        left_raw = self.left_raw
        right_raw = self.right_raw

        self.events.add(left_raw, AND)
        self.events.add(right_raw, AND)

        return True



import tensorflow as tf

from neupy.core.properties import ProperFractionProperty, NumberProperty
from .base import BaseLayer


__all__ = ('Dropout', 'GaussianNoise')


class Dropout(BaseLayer):
    """
    Dropout layer

    Parameters
    ----------
    proba : float
        Fraction of the input units to drop. Value needs to be
        between ``0`` and ``1``.

    {BaseLayer.Parameters}

    Methods
    -------
    {BaseLayer.Methods}

    Attributes
    ----------
    {BaseLayer.Attributes}
    """
    proba = ProperFractionProperty(required=True)

    def __init__(self, proba, **options):
        super(Dropout, self).__init__(proba=proba, **options)

    def output(self, input_value):
        if not self.training_state:
            return input_value
        return tf.nn.dropout(input_value, keep_prob=(1.0 - self.proba))

    def __repr__(self):
        classname = self.__class__.__name__
        return "{}(proba={})".format(classname, self.proba)


class GaussianNoise(BaseLayer):
    """
    Add gaussian noise to the input value. Mean and standard
    deviation are layer's parameters.

    Parameters
    ----------
    std : float
        Standard deviation of the gaussian noise. Values needs to
        be greater than zero. Defaults to ``1``.

    mean : float
        Mean of the gaussian noise. Defaults to ``0``.

    {BaseLayer.Parameters}

    Methods
    -------
    {BaseLayer.Methods}

    Attributes
    ----------
    {BaseLayer.Attributes}
    """
    std = NumberProperty(default=1, minval=0)
    mean = NumberProperty(default=0)

    def __init__(self, mean=1, std=0, **options):
        super(GaussianNoise, self).__init__(mean=mean, std=std, **options)

    def output(self, input_value):
        if not self.training_state:
            return input_value

        noise = tf.random_normal(
            shape=tf.shape(input_value),
            mean=self.mean,
            stddev=self.std)

        return input_value + noise

    def __repr__(self):
        classname = self.__class__.__name__
        return "{}(mean={}, std={})".format(classname, self.mean, self.std)



import numpy as np
import tensorflow as tf
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from sklearn import model_selection, datasets

from neupy import layers, environment, algorithms
from neupy.exceptions import LayerConnectionError


environment.reproducible()


def load_data():
    mnist = datasets.fetch_mldata('MNIST original')

    data = mnist.data / 255.
    data = data - data.mean(axis=0)

    x_train, x_test = model_selection.train_test_split(
        data.astype(np.float32), test_size=(1 / 7.)
    )
    return x_train, x_test


def generate_and_plot_sampels(network, nx, ny):
    space_x = np.linspace(-8, 8, nx)
    space_y = np.linspace(-8, 8, ny)

    plt.figure(figsize=(12, 12))

    grid = gridspec.GridSpec(nx, ny)
    grid.update(wspace=0, hspace=0)

    for i, x in enumerate(space_x):
        for j, y in enumerate(space_y):
            sample = np.array([x, y])
            digit = generator.predict(sample)
            image = digit[0].reshape((28, 28))

            plt.subplot(grid[nx * i + j])
            plt.imshow(image, cmap='Greys_r')
            plt.axis('off')

    plt.show()


class GaussianSample(layers.BaseLayer):
    def validate(self, input_shapes):
        classname = self.__class__.__name__

        if not isinstance(input_shapes, list):
            raise LayerConnectionError(
                "{} layer expected 2 inputs, got 1"
                "".format(classname))

        if len(input_shapes) != 2:
            n_inputs = len(input_shapes)
            raise LayerConnectionError(
                "{} layer expected 2 inputs, got {}"
                "".format(classname, n_inputs))

        for input_shape in input_shapes:
            ndim = len(input_shape)

            if ndim != 1:
                raise LayerConnectionError(
                    "Input layer to {} should be 2D, "
                    "got {}D".format(classname, ndim))

    @property
    def output_shape(self):
        if self.input_shape:
            return self.input_shape[0]

    def output(self, *input_values):
        mu, sigma = input_values
        return mu + tf.exp(sigma) * tf.random_normal(tf.shape(mu))


def binary_crossentropy(expected, predicted):
    epsilon = 1e-7  # smallest positive 32-bit float number
    predicted = tf.clip_by_value(predicted, epsilon, 1.0 - epsilon)

    return -tf.reduce_sum(
        expected * tf.log(predicted) +
        (1 - expected) * tf.log(1 - predicted),
        axis=1
    )


def vae_loss(expected, predicted):
    graph = tf.get_default_graph()
    x = graph.get_tensor_by_name('network-input/to-layer-input:0')

    mean = (encoder > mu).output(x)
    log_var = (encoder > sigma).output(x)

    epsilon = 1e-7
    predicted = tf.clip_by_value(predicted, epsilon, 1.0 - epsilon)

    crossentropy_loss = binary_crossentropy(expected, predicted)
    kl_loss = tf.reduce_sum(
        1 + 2 * log_var - tf.square(mean) - tf.exp(2 * log_var),
        axis=1
    )
    return tf.reduce_mean(crossentropy_loss - 0.5 * kl_loss)


# Construct Variational Autoencoder
encoder = layers.Input(784, name='input') > layers.Tanh(256)

# Two is the maximum number of dimensions that we can visualize
mu = layers.Linear(2, name='mu')
sigma = layers.Linear(2, name='sigma')
sampler = [mu, sigma] > GaussianSample()

decoder = layers.Tanh(256) > layers.Sigmoid(784)

# Train network
network = algorithms.RMSProp(
    [
        encoder,
        sampler,
        decoder,
    ],

    error=vae_loss,
    batch_size=128,
    shuffle_data=True,
    step=0.001,
    verbose=True,

    decay_rate=0.01,
    addons=[algorithms.WeightDecay],
)

x_train, x_test = load_data()
network.train(x_train, x_train, x_test, x_test, epochs=50)

# Sample digits from the obtained distribution
generator = algorithms.GradientDescent(layers.Input(2) > decoder)
generate_and_plot_sampels(generator, 25, 25)
